from whad.dot15d4.stack.manager import Dot15d4Manager
from whad.dot15d4.stack.service import Dot15d4Service
from whad.dot15d4.stack.mac.database import MACPIB
from whad.dot15d4.stack.mac.exceptions import MACTimeoutException, MACAssociationFailure
from whad.dot15d4.stack.mac.helpers import is_short_address
from whad.dot15d4.stack.mac.constants import MACScanType, MACConstants, MACAddressMode, \
    MACDeviceType, MACPowerSource, MACBeaconType, MACAssociationStatus
from whad.dot15d4.stack.mac.network import Dot15d4PANNetwork
from whad.dot15d4.stack.mac.energy import EDMeasurement
from whad.common.stack import Layer, alias, source, state
from scapy.layers.dot15d4 import Dot15d4Data, Dot15d4Beacon, Dot15d4Cmd, \
    Dot15d4Ack, Dot15d4, Dot15d4CmdAssocReq, Dot15d4CmdAssocResp
from whad.exceptions import RequiredImplementation

from threading import Thread
from time import time, sleep
from queue import Queue, Empty

import logging

logger = logging.getLogger(__name__)



class MACService(Dot15d4Service):
    """
    This class represents a MAC service, exposing a standardized API.
    """
    def __init__(self, manager, name=None):
        super().__init__(
            manager,
            name=name,
            timeout_exception_class=MACTimeoutException
        )

class MACDataService(MACService):
    """
    MAC service processing Data packets.
    """
    def __init__(self, manager):
        super().__init__(manager, name="mac_data")


    @Dot15d4Service.request("MCPS-DATA")
    def data(
                self,
                msdu,
                msdu_handle=0,
                source_address_mode=MACAddressMode.SHORT,
                destination_pan_id=0xFFFF,
                destination_address=0xFFFF,
                destination_address_mode=MACAddressMode.SHORT,
                pan_id_suppressed=False,
                sequence_number_suppressed=False,
                wait_for_ack=False
    ):

        data = Dot15d4Data()
        if destination_pan_id is not None:
            data.dest_panid = destination_pan_id
        if destination_address is not None:
            data.dest_addr = destination_address
        if not pan_id_suppressed:
            data.src_panid = self.database.get("macPanId")

        if source_address_mode == MACAddressMode.SHORT:
            data.src_addr = self.database.get("macShortAddress")
        elif source_address_mode == MACAddressMode.EXTENDED:
            data.src_addr = self.database.get("macExtendedAddress")
        data = data/msdu

        while not self.manager.all_pending_transactions_processed(destination_address):
            sleep(0.1)

        ack = self.manager.send_data(
                                        data,
                                        wait_for_ack=wait_for_ack,
                                        source_address_mode=source_address_mode,
                                        destination_address_mode=destination_address_mode
        )
        return ack


    def on_data_pdu(self, pdu):
        self.indicate_data(pdu)

    def on_ack_pdu(self, pdu):
        pass


    @Dot15d4Service.indication("MCPS-DATA")
    def indicate_data(self, pdu):
        source_pan_id = pdu.src_panid if hasattr(pdu, "src_panid") else None
        source_address = pdu.src_addr if hasattr(pdu, "src_addr") else None
        destination_pan_id = pdu.dest_panid if hasattr(pdu, "dest_panid") else None
        destination_address = pdu.dest_addr if hasattr(pdu, "dest_addr") else None
        payload = pdu[Dot15d4Data].payload if Dot15d4Data in pdu else pdu[Dot15d4].payload
        link_quality = pdu.metadata.lqi if hasattr(pdu, "metadata") and hasattr(pdu.metadata, "lqi") else 255
        return (payload, {
            "destination_pan_id":destination_pan_id,
            "destination_address":destination_address,
            "source_pan_id":source_pan_id,
            "source_address":source_address,
            "link_quality":link_quality
        })


class MACManagementService(MACService):
    """
    MAC service processing Management packets.
    """
    def __init__(self, manager):
        super().__init__(manager, name="mac_management")
        self._beaconing_thread = None
        self._beaconing_thread_enabled = False
        self._samples_queue = Queue()


    def add_ed_sample_to_queue(self, ed_sample):
        """
        Add an ED sample to the queue for later processing.
        """
        self._samples_queue.put(ed_sample)

    # Requests
    @Dot15d4Service.request("MLME-GET")
    def get(self, attribute):
        """
        Implement the MLME-GET request operation.
        """
        return self.database.get(attribute)

    @Dot15d4Service.request("MLME-SET")
    def set(self, attribute, value):
        """
        Implement the MLME-SET request operation.
        """
        return self.database.set(attribute, value)

    @Dot15d4Service.request("MLME-RESET")
    def reset(self, set_default_pib=True):
        """
        Implement the MLME-RESET request operation.
        """
        if set_default_pib:
            self.database.reset()
            self.manager.set_extended_address(self.database.get("macExtendedAddress"))
            return True
        return False


    @Dot15d4Service.request("MLME-SYNC")
    def sync(
                self,
                channel_page=0,
                channel=11,
                track_beacon=False
    ):
        """
        Implement the MLME-SYNC request operation.
        """
        self.manager.set_channel_page(channel_page)
        self.manager.set_channel(channel)

        # we don't do anything according to track beacon parameter,
        # because our radio will always remain enabled.
        return True


    @Dot15d4Service.request("MLME-POLL")
    def poll(self, coordinator_pan_id=0, coordinator_address=0):
        """
        Implement the MLME-POLL request operation.
        """
        ack = self.manager.send_data(
            Dot15d4Cmd(
                cmd_id = "DataReq",
                dest_addr = coordinator_address,
                dest_panid = coordinator_pan_id,
                src_addr = self.database.get("macShortAddress"),
                src_panid = self.database.get("macPanId")
            ),
            wait_for_ack=True,
            return_ack=True,
            source_address_mode=MACAddressMode.SHORT,
            destination_address_mode=MACAddressMode.SHORT,
        )
        if ack is None:
            return False
        if ack.fcf_pending:
            try:
                data = self.manager.get_service("data").wait_for_packet(
                    lambda pkt: Dot15d4Data in pkt or pkt.fcf_frametype == 1
                )
                return True
            except MACTimeoutException:
                return False
        return False


    @Dot15d4Service.request("MLME-BEACON")
    def beacon(
                    self,
                    beacon_type=MACBeaconType.BEACON,
                    channel_page=0,
                    channel=11,
                    superframe_order=15,
                    header_ie_list=[],
                    payload_ie_list=[],
                    header_ie_id_list=[],
                    nested_ie_sub_id_list=[],
                    beacon_security_level=0,
                    beacon_key_id_mode=0,
                    beacon_key_source=b"",
                    beacon_key_index=0,
                    source_address_mode=MACAddressMode.SHORT,
                    destination_address_mode=MACAddressMode.NONE,
                    destination_address=0xFFFF,
                    bsn_suppression=False
    ):
        if beacon_type == MACBeaconType.ENHANCED_BEACON:
            raise RequiredImplementation("EnhancedBeacon")

        src_panid = self.database.get("macPanId")
        src_addr = (
            self.database.get("macShortAddress") if
            source_address_mode == MACAddressMode.SHORT else
            self.database.get("macExtendedAddress")
        )

        sf_beaconorder = self.database.get("macBeaconOrder")
        sf_assocpermit = self.database.get("macAssociationPermit")
        is_coordinator = (
            self.database.get("macCoordShortAddress") == self.database.get("macShortAddress")
        )
        sf_battlifeextend = self.database.get("macBattLifeExt")
        beacon_payload = self.database.get("macBeaconPayload")

        beacon = Dot15d4Beacon(
            src_panid=src_panid,
            src_addr=src_addr,
            sf_sforder=superframe_order,
            sf_beaconorder=sf_beaconorder,
            sf_assocpermit=sf_assocpermit,
            sf_pancoord=is_coordinator,
            sf_battlifeextend=sf_battlifeextend
        ) / beacon_payload

        self.manager.send_beacon(
            beacon,
            source_address_mode=source_address_mode,
            destination_address_mode=destination_address_mode
        )
        return True

    @Dot15d4Service.request("MLME-START")
    def start(
                    self,
                    pan_id,
                    channel_page=0,
                    channel=11,
                    start_time=0,
                    beacon_order=15,
                    superframe_order=15,
                    pan_coordinator=True,
                    battery_life_extension=False,
                    coord_realignement=False,
                    coord_realign_security_level=0,
                    coord_realign_key_id_mode=0,
                    coord_realign_key_index=0,
                    coord_realign_key_source=b"",
                    beacon_security_level=0,
                    beacon_key_id_mode=0,
                    beacon_key_source=b"",
                    beacon_key_index=0,
                    header_ie_list=[],
                    payload_ie_list=[],
                    header_ie_id_list=[],
                    nested_ie_sub_id_list=[]
    ):
        """
        Implement the MLME-START request operation.
        """
        if coord_realignement:
            raise RequiredImplementation("CoordinatorRealignment")
        else:
            self.database.set("macBeaconOrder", 3)
            self.database.set("macSuperframeOrder", 1)
            self.database.set("macPanId", pan_id)
            self.manager.get_layer('phy').set_channel_page(channel_page)
            self.manager.get_layer('phy').set_channel(channel)


            src_panid = self.database.get("macPanId")
            src_addr = self.database.get("macShortAddress")

            sf_beaconorder = self.database.get("macBeaconOrder")
            sf_assocpermit = self.database.get("macAssociationPermit")
            is_coordinator = (
                self.database.get("macCoordShortAddress") == self.database.get("macShortAddress")
            )
            sf_battlifeextend = self.database.get("macBattLifeExt")
            beacon_payload = self.database.get("macBeaconPayload")

            beacon = Dot15d4Beacon(
                src_panid=src_panid,
                src_addr=src_addr,
                sf_sforder=superframe_order,
                sf_beaconorder=sf_beaconorder,
                sf_assocpermit=sf_assocpermit,
                sf_pancoord=is_coordinator,
                sf_battlifeextend=sf_battlifeextend
            ) / beacon_payload

            self.database.set("macLastBeacon", Dot15d4() / beacon)

            # if beaconOrder < 15, we start a beacon-enabled network
            if beacon_order < 15:
                self.database.set("macBattLifeExt", battery_life_extension)
                if not pan_coordinator:
                    # Start beaconing after start time
                    self._start_beaconing(start_time)
                else:
                    # Start transmitting immediatly
                    self._start_beaconing()

    @Dot15d4Service.response("MLME-ASSOCIATE")
    def associate_response(
        self,
        device_address,
        assoc_short_address,
        security_level=0,
        key_id_mode=0,
        key_source=0,
        key_index=0,
        channel_offset=0,
        hopping_sequence=b"",
        dsme_association=False,
        allocation_order=0,
        bi_index=0,
        superframe_id=0,
        slot_id=0,
        channel_index=0,
        association_status=MACAssociationStatus.ASSOCIATION_SUCCESSFUL
    ):

        association_response = Dot15d4Cmd(
            cmd_id="AssocResp",
            dest_addr = device_address,
            src_addr = self.database.get("macExtendedAddress"),
            dest_panid = self.database.get("macPanId"),

        ) / Dot15d4CmdAssocResp(
            short_address = assoc_short_address,
            association_status = int(association_status)
        )

        self.manager.add_pending_transaction(
            association_response,
            source_address_mode=MACAddressMode.EXTENDED,
            destination_address_mode=MACAddressMode.EXTENDED
        )

        # dirty, but async issue when pending transaction in use
        '''
        sleep(2)
        self.manager.send_data(
            association_response,
            source_address_mode=MACAddressMode.EXTENDED,
            destination_address_mode=MACAddressMode.EXTENDED,
        )
        '''

        return True

    @Dot15d4Service.request("MLME-ASSOCIATE")
    def associate(
                    self,
                    channel_page=0,
                    channel=11,
                    coordinator_pan_id=0,
                    coordinator_address=0,
                    device_type=MACDeviceType.FFD,
                    power_source=MACPowerSource.ALTERNATING_CURRENT_SOURCE,
                    idle_receiving=True,
                    allocate_address=True,
                    security_capability=False,
                    fast_association=False
    ):
        """
        Implement the MLME-ASSOCIATE request operation.
        """

        logger.debug("[mac] starting association with %s", coordinator_pan_id)
        self.database.set("macPanId", coordinator_pan_id)
        if is_short_address(coordinator_address):
            self.database.set("macCoordShortAddress", coordinator_address)
        else:
            self.database.set("macCoordExtendedAddress", coordinator_address)
        self.manager.set_channel_page(channel_page)
        self.manager.set_channel(channel)

        duration = (
            self.database.get("macResponseWaitTime") *
            MACConstants.A_BASE_SUPERFRAME_DURATION *
            self.manager.get_layer('phy').symbol_duration
        )

        try:
            logger.debug("[mac] sending association request ...")
            acked = self.manager.send_data(
                    Dot15d4Cmd(
                        cmd_id = "AssocReq",
                        dest_addr = coordinator_address,
                        dest_panid = coordinator_pan_id,
                        src_addr = self.database.get("macExtendedAddress"),
                        src_panid = 0xFFFF
                    )/
                    Dot15d4CmdAssocReq(
        				allocate_address = int(allocate_address),
        				security_capability = int(security_capability),
        				power_source = int(power_source == MACPowerSource.ALTERNATING_CURRENT_SOURCE),
        				device_type = int(device_type == MACDeviceType.FFD),
        				receiver_on_when_idle = int(idle_receiving),
        				alternate_pan_coordinator = int(fast_association)
        		  )
                ,
                wait_for_ack=True,
                source_address_mode=MACAddressMode.EXTENDED
            )

            if acked:
                logger.debug("[mac] association request sent, send data request ...")
                sleep(duration/1000000)
                ack = self.manager.send_data(
                    Dot15d4Cmd(
                        cmd_id = "DataReq",
                        dest_addr = coordinator_address,
                        dest_panid = coordinator_pan_id,
                        src_addr = self.database.get("macExtendedAddress"),
                        src_panid = 0xFFFF
                    ),
                    wait_for_ack=True,
                    source_address_mode=MACAddressMode.EXTENDED
                )

                if acked:
                    logger.debug("[mac] data request sent, waiting for association response ...")
                    try:
                        association_response = self.wait_for_packet(
                            lambda pkt: Dot15d4CmdAssocResp in pkt, timeout=5.0
                        )
                        if association_response.association_status == 0:
                            self.database.set("macRxOnWhenIdle", idle_receiving)
                            self.manager.set_short_address(association_response.short_address)
                            if association_response.fcf_srcaddrmode == 2:
                                self.database.set("macCoordShortAddress", association_response.src_addr)
                            elif association_response.fcf_srcaddrmode == 3:
                                self.database.set("macCoordExtendedAddress", association_response.src_addr)

                            return True
                        else:
                            logger.debug("[mac] associaton unsucessful")
                            raise MACAssociationFailure("association response unsuccessful (status={})".format(hex(association_response.association_status)))
                    except MACTimeoutException:
                        logger.debug("[mac] association timeout")
                        raise MACAssociationFailure("association response timeout. ")
                else:
                    logger.debug("[mac] no ack for data request")
                    raise MACAssociationFailure("no acknowledgement received for DataRequest. ")
            else:
                logger.debug("[mac] no ack for association request")
                raise MACAssociationFailure("no acknowledgement received for AssociationRequest. ")

        except MACAssociationFailure as err:
            logger.info("[{}] Association failure - {}. ".format(self._name, err.reason))
            self.database.set("macPanId", 0xFFFF)
            return False


    @Dot15d4Service.request("MLME-SCAN")
    def scan(
                self,
                scan_type=MACScanType.ACTIVE,
                channel_page=0,
                scan_channels=0x7fff800,
                scan_duration=5
    ):
        """
        Implement the MLME-SCAN request operation.
        """
        # Convert channel map to channels list
        if isinstance(scan_channels, int):
            channels = []
            for i in range(0,27):
                if scan_channels & (1 << i) != 0:
                    channels.append(i)
        elif isinstance(scan_channels, list):
            channels = scan_channels
        else:
            return False

        # Check scan_duration validity
        if scan_duration < 0 or scan_duration > 14:
            return False

        # Convert scan duration
        duration = (
            MACConstants.A_BASE_SUPERFRAME_DURATION *
            ((2**scan_duration) + 1) *
            self.manager.get_layer('phy').symbol_duration
        )

        # Select the right scan
        if scan_type == MACScanType.ACTIVE:
            return self._perform_legacy_scan(channel_page, channels, duration, active=True)
        elif scan_type == MACScanType.PASSIVE:
            return self._perform_legacy_scan(channel_page, channels, duration, active=False)
        elif scan_type == MACScanType.ENERGY_DETECTION:
            return self._perform_ed_scan(channel_page, channels, duration)
        else:
            raise RequiredImplementation("Scan")

    # Indications
    @Dot15d4Service.indication("MLME-BEACON-NOTIFY")
    def indicate_beacon_notify(self, pan_descriptor, beacon_payload):
        return (beacon_payload, {
                    "pan_descriptor": pan_descriptor,
        })

    @Dot15d4Service.indication("MLME-ASSOCIATE")
    def indicate_associate(self, pdu):
        """
        Implements the MLME-ASSOCIATE indication operation.
        """
        source_address = pdu.src_addr if hasattr(pdu, "src_addr") else None
        capability_information = (
            0 |
            (int(pdu.device_type) << 1) |
            (int(pdu.power_source) << 2) |
            (int(pdu.receiver_on_when_idle) << 3) |
            (int(pdu.alternate_pan_coordinator) << 4) |
            (int(pdu.security_capability) << 6) |
            (int(pdu.allocate_address) << 7)
        )
        security_level = (
            pdu.sec_sc_seclevel if
            hasattr(pdu, "sec_sc_seclevel") and
            hasattr(pdu, "fcf_security") and
            pdu.fcf_security else
            0
        )
        key_id_mode = (
            pdu.sec_sc_keyidmode if
            hasattr(pdu, "sec_sc_keyidmode") and
            hasattr(pdu, "fcf_security") and
            pdu.fcf_security else
            0
        )
        key_source = (
            pdu.sec_keyid_keysource if
            hasattr(pdu, "sec_keyid_keysource") and
            hasattr(pdu, "fcf_security") and
            pdu.fcf_security else
            0
        )
        key_index = (
            pdu.sec_keyid_keyindex if
            hasattr(pdu, "sec_keyid_keyindex") and
            hasattr(pdu, "fcf_security") and
            pdu.fcf_security else
            0
        )
        # Default values here
        channel_offset = 0
        hopping_sequence_id = 0
        dsme_association = False
        direction = 0
        allocation_order = 0
        hopping_sequence_request = False

        return (
            source_address,
            {
                "capability_information":capability_information,
                "security_level":security_level,
                "key_id_mode":key_id_mode,
                "key_source":key_source,
                "key_index":key_index,
                "channel_offset":channel_offset,
                "hopping_sequence_id": hopping_sequence_id,
                "dsme_association":dsme_association,
                "direction":direction,
                "allocation_order": allocation_order,
                "hopping_sequence_request":hopping_sequence_request
            }
        )
    @Dot15d4Service.indication("MLME-BEACON-REQUEST")
    def indicate_beacon_request(self, pdu):
        """
        Implements the MLME-BEACON-REQUEST indication operation.
        """
        if pdu.fcf_srcaddrmode == 0:
            source_address_mode = MACAddressMode.NONE
        elif pdu.fcf_srcaddrmode == 2:
            source_address_mode = MACAddressMode.SHORT
        elif pdu.fcf_srcaddrmode == 3:
            source_address_mode = MACAddressMode.EXTENDED
        else:
            source_address_mode = MACAddressMode.NONE

        source_address = pdu.src_addr if hasattr(pdu, "src_addr") else None
        destination_pan_id = pdu.dest_panid if hasattr(pdu, "dest_panid") else 0xFFFF

        # Not processed for now, default values
        beacon_type = MACBeaconType.BEACON
        header_ie_list = []
        payload_ie_list = []
        print((beacon_type, {
                    "source_address_mode": source_address_mode,
                    "source_address": source_address,
                    "destination_pan_id": destination_pan_id,
                    "header_ie_list": header_ie_list,
                    "payload_ie_list": payload_ie_list
        })
        )
        return (beacon_type, {
                    "source_address_mode": source_address_mode,
                    "source_address": source_address,
                    "destination_pan_id": destination_pan_id,
                    "header_ie_list": header_ie_list,
                    "payload_ie_list": payload_ie_list
        })

    # Low level primitives and helpers
    def _perform_legacy_scan(self, channel_page, channels, duration, active=True):
        pan_descriptors = []

        # Enter MAC Promiscuous mode to prevent filtering
        oldPromiscuousMode = self.database.get("macPromiscuousMode")
        self.database.set("macPromiscuousMode", True)
        for channel in channels:
            self.manager.set_channel_page(channel_page)
            self.manager.set_channel(channel)
            #print("Channel", channel)
            start_time = time()
            if active:
                self.manager.send_data(
                    Dot15d4Cmd(
                        cmd_id="BeaconReq",
                        dest_addr=0xFFFF,
                        dest_panid=0xFFFF
                    )
                )

            while (time() - start_time) < duration:
                try:
                    beacon = self.wait_for_packet(
                        lambda pkt: Dot15d4Beacon in pkt, timeout=0.0001
                    )
                    pan_descriptor = Dot15d4PANNetwork(beacon, channel_page, channel)
                    self.indicate_beacon_notify(pan_descriptor, beacon[Dot15d4Beacon].payload)
                    if pan_descriptor not in pan_descriptors:
                        pan_descriptors.append(pan_descriptor)
                except MACTimeoutException:
                    pass

        self.database.set("macPromiscuousMode", oldPromiscuousMode)
        return pan_descriptors


    def _perform_ed_scan(self, channel_page, channels, duration):
        ed_measurements = []

        for channel in channels:
            self.manager.set_channel_page(channel_page)
            self._samples_queue.queue.clear()
            self.manager.perform_ed_scan(channel)

            start_time = time()
            samples = []
            while (time() - start_time) < duration:
                try:
                    sample = self._samples_queue.get(block=False)
                    samples.append(sample)
                except Empty:
                    pass
            ed_measurements.append(EDMeasurement(samples, channel_page, channel))
        return ed_measurements

    def _start_beaconing(self, start_time=0):
        # Check if we are beacon-enabled PAN
        if self.database.get("macBeaconOrder") == 15:
            # non beacon-enabled, terminate
            return False

        beacon_interval = (
            MACConstants.A_BASE_SUPERFRAME_DURATION *
            (2**self.database.get("macBeaconOrder"))
        )

        if self.database.get("macSuperframeOrder") == 15:
            superframe = False
        else:
            superframe = True

        superframe_duration = (
            MACConstants.A_BASE_SUPERFRAME_DURATION *
            (2**self.database.get("macSuperframeOrder"))
        )

        if self._beaconing_thread is not None:
                self._stop_beaconing()

        # Waiting time before start
        init_time = time()
        while (time() - init_time) < start_time:
            pass

        self._beaconing_thread_enabled = True
        self._beaconing_thread = Thread(target=self._beaconing,args=(beacon_interval, superframe, superframe_duration), daemon=True)
        self._beaconing_thread.start()

    def _stop_beaconing(self):
        self._beaconing_thread_enabled = False
        self._beaconing_thread = None

    def _beaconing(self, beacon_interval, superframe, superframe_duration):
        """
        Thread responsible of transmitting beacon and managing superframe.
        """
        last_beacon_time = None
        in_superframe = False

        while self._beaconing_thread_enabled:
            current_time = time() * 1000
            if last_beacon_time is None or (current_time - last_beacon_time) >= beacon_interval:
                last_beacon_time = current_time
                in_superframe = True
                self.beacon()
                print("superframe opened")
            if superframe and in_superframe and (current_time - last_beacon_time) >= superframe_duration:
                in_superframe = False
                print("superframe closed")
            sleep(0.01)

    # Input callbacks
    def on_cmd_pdu(self, pdu):
        if pdu.cmd_id == 1: # Association Request
            self.indicate_associate(pdu)
        elif pdu.cmd_id == 4:
            self.on_data_request(pdu)
            self.indicate_data_req(pdu)
        elif pdu.cmd_id == 7: # Beacon Request
            if self.database.get("macBeaconAutoRespond"):
                self.beacon()
            else:
                self.indicate_beacon_request(pdu)
        self.add_packet_to_queue(pdu)

    @Dot15d4Service.indication("MLME−DATA-REQ")
    def indicate_data_req(self, pdu):
        """
        Pseudo indication for Data Request.
        """
        return (pdu, {})

    def on_data_request(self, pdu):
        """
        Callback triggered when a data request is received.

        This method will check if some pending data are available and transmit it if necessary.
        """
        pending = self.manager.get_pending_transaction(pdu.src_addr)
        if pending is not None:
            packet, source_address_mode, destination_address_mode = pending

            self.manager.send_data(
                packet,
                source_address_mode=source_address_mode,
                destination_address_mode=destination_address_mode
            )


    def on_beacon_pdu(self, pdu):
        """
        Callback triggered when a beacon is received.
        """
        self.add_packet_to_queue(pdu)
        macAutoRequest = self.database.get("macAutoRequest")
        if macAutoRequest:
            macPanId = self.database.get("macPanId")
            if pdu.src_panid == macPanId:
                # Check if some data are pending for us, and transmit a Data Request if necessary.
                extendedAddress = self.database.get("macExtendedAddress")
                shortAddress = self.database.get("macShortAddress")
                if (
                        extendedAddress in pdu.pa_long_addresses or
                        shortAddress in pdu.pa_short_addresses
                ):
                    self.poll(coordinator_pan_id=pdu.src_panid, coordinator_address=pdu.src_addr)

    def on_ed_sample(self, timestamp, sample):
        self.add_ed_sample_to_queue(sample)


@state(MACPIB)
@alias('mac')
class MACManager(Dot15d4Manager):
    """
    This class implements the Medium Access Control Manager (MAC - defined in 802.15.4 specification).

    The Medium Access Control manager handles all the low-level operations:
    - 802.15.4 management control (handles beaconing, frame validation, timeslots and associations)
    - 802.15.4 data (forward to upper layer, i.e. NWK)

    The API is exposed by two services.
    """
    def init(self):
        self.add_service("data", MACDataService(self))
        self.add_service("management", MACManagementService(self))
        self.__ack_queue = Queue()
        self.__pending_transactions = {}
        # Move it to connector ?
        #self.set_extended_address(self.database.get("macExtendedAddress"))


    def all_pending_transactions_processed(self, address):
        if address in self.__pending_transactions:
            return self.__pending_transactions[address].empty()
        else:
            return True

    def add_pending_transaction(self, packet, source_address_mode=None, destination_address_mode=None):
        """
        Put outgoing data as pending.
        """
        if packet.dest_addr not in self.__pending_transactions:
            self.__pending_transactions[packet.dest_addr] = Queue()

        self.__pending_transactions[packet.dest_addr].put(
            (packet, source_address_mode, destination_address_mode)
        )

    def get_pending_transaction(self, address):
        """
        Get the latest pending transaction available.

        Return None if transaction queue is empty.
        """
        if (
            address in self.__pending_transactions and
            not self.__pending_transactions[address].empty()
        ):
            return self.__pending_transactions[address].get()
        else:
            return None

    @source('phy', 'energy_detection')
    def on_ed_sample(self, sample, timestamp):
        self.get_service("management").on_ed_sample(timestamp, sample)

    @source('phy', 'pdu')
    def on_pdu(self, pdu):
        if self.match_filter(pdu):
            if pdu.fcf_frametype == 0x01 or Dot15d4Data in pdu:
                self.get_service("data").on_data_pdu(pdu)
            elif pdu.fcf_frametype == 0x02 or Dot15d4Ack in pdu:
                self.__ack_queue.put(pdu, block=True, timeout=None)
                self.get_service("data").on_ack_pdu(pdu)
            elif pdu.fcf_frametype == 0x03 or Dot15d4Cmd in pdu:
                self.get_service("management").on_cmd_pdu(pdu)
            elif pdu.fcf_frametype == 0x00 or Dot15d4Beacon in pdu:
                self.get_service("management").on_beacon_pdu(pdu)
            else:
                logger.warning("[mac_manager] Malformed PDU received: %s", repr(pdu))

    def match_filter(self, pdu):
        macPromiscuousMode = self.database.get("macPromiscuousMode")
        if macPromiscuousMode:
            return True

        macImplicitBroadcast = self.database.get("macImplicitBroadcast")
        if (
                not hasattr(pdu, "dest_panid") and
                not hasattr(pdu, "dest_addr") and
                macImplicitBroadcast
        ):
            return True

        macPanId = self.database.get("macPanId")
        if (
                not hasattr(pdu, "dest_panid") and
                not hasattr(pdu, "dest_addr") and
                hasattr(pdu, "src_panid") and
                pdu.src_panid != macPanId
        ):
            return False

        if hasattr(pdu, "dest_panid"):
            if pdu.dest_panid != macPanId and pdu.dest_panid != 0xFFFF:
                return False


        if hasattr(pdu, "dest_addr"):
            macShortAddress = self.database.get("macShortAddress")
            macExtendedAddress = self.database.get("macExtendedAddress")
            if (
                pdu.dest_addr != macShortAddress and
                pdu.dest_addr != macExtendedAddress and
                pdu.dest_addr != 0xFFFF
            ):
                return False

        return True


    def wait_for_ack(self, timeout=None):
        """Wait for a ACK or error.

        :param float timeout: Timeout value (default: 30 seconds)
        """
        if timeout is None:
            timeout = self.database.get("macAckTimeout")
        start_time = time()
        while (time() - start_time) < timeout:
            try:
                msg = self.__ack_queue.get(block=False,timeout=0.1)
                return msg
            except Empty:
                pass
        raise MACTimeoutException

    def _choose_pan_id_compression(self, packet, destination_address_mode, source_address_mode):
        if packet.fcf_framever in (0, 1):
            if (
                destination_address_mode != MACAddressMode.NONE and
                source_address_mode != MACAddressMode.NONE
            ):

                if (
                    hasattr(packet, "dest_panid") and
                    hasattr(packet, "src_panid") and
                    packet.dest_panid == packet.src_panid
                ):
                    packet.fcf_panidcompress = 1
                else:
                    packet.fcf_panidcompress = 0
            else:
                packet.fcf_panidcompress = 0
        elif packet.fcf_framever == 2:
            framecompress = PANID_COMPRESSION_TABLE[
                (
                    destination_address_mode,
                    source_address_mode,
                    hasattr(packet, "dest_panid"),
                    hasattr(packet, "src_panid")
                )
            ]
            if (
                (
                    source_address_mode == MACAddressMode.SHORT and
                    destination_address_mode != MACAddressMode.NONE
                ) or
                (
                    destination_address_mode == MACAddressMode.SHORT and
                    source_address_mode != MACAddressMode.NONE
                )
            ):
                if (
                    hasattr(packet, "dest_panid") and
                    hasattr(packet, "src_panid") and
                    packet.dest_panid == packet.src_panid
                ):
                    packet.fcf_panidcompress = 1
                else:
                    packet.fcf_panidcompress = 0
            else:
                packet.fcf_panidcompress = framecompress
        else:
            packet.fcf_panidcompress = 0
        return packet

    def send_data(self, packet, wait_for_ack=False, return_ack=False, source_address_mode=None, destination_address_mode=None):
        if source_address_mode is not None:
            if source_address_mode == MACAddressMode.NONE:
                fcf_srcaddrmode = 0
            elif source_address_mode == MACAddressMode.SHORT:
                fcf_srcaddrmode = 2
            else:
                fcf_srcaddrmode = 3
        else:
            source_address_mode = MACAddressMode.NONE
            fcf_srcaddrmode = 0

        if destination_address_mode is not None:
            if destination_address_mode == MACAddressMode.NONE:
                fcf_destaddrmode = 0
            elif destination_address_mode == MACAddressMode.SHORT:
                fcf_destaddrmode = 2
            else:
                fcf_destaddrmode = 3
        else:
            destination_address_mode = MACAddressMode.SHORT
            fcf_destaddrmode = 2

        packet = Dot15d4(
                    fcf_srcaddrmode=fcf_srcaddrmode,
                    fcf_destaddrmode=fcf_destaddrmode
        )/packet

        if wait_for_ack:
            packet.fcf_ackreq = 1

        # Build PAN ID compression field
        packet = self._choose_pan_id_compression(packet, destination_address_mode, source_address_mode)

        #    packet.fcf_panidcompress = 1

        sequence_number = self.database.get("macDataSequenceNumber")
        packet.seqnum = sequence_number
        self.database.set("macDataSequenceNumber", (sequence_number + 1) % 256)
        self.send('phy', packet, tag='pdu')
        wait_counter = 5
        if wait_for_ack:
            ack = None
            while ack is None or ack.seqnum != sequence_number:
                try:
                    ack = self.wait_for_ack()
                except MACTimeoutException:
                    wait_counter = wait_counter - 1
                    if wait_counter <= 0:
                        if return_ack:
                            return None
                        return False
            if return_ack:
                return ack
            return True
        else:
            return True

    def set_short_address(self, address):
        self.database.set("macShortAddress", address)
        self.get_layer('phy').set_short_address(address)

    def set_extended_address(self, address):
        self.database.set("macExtendedAddress", address)
        self.get_layer('phy').set_extended_address(address)

    def set_channel_page(self, page):
        self.get_layer('phy').set_channel_page(page)

    def set_channel(self, channel):
        self.get_layer('phy').set_channel(channel)

    def perform_ed_scan(self, channel):
        self.get_layer('phy').perform_ed_scan(channel)

    def send_beacon(self, packet, source_address_mode=None, destination_address_mode=None):
        if source_address_mode is not None:
            if source_address_mode == MACAddressMode.NONE:
                fcf_srcaddrmode = 0
            elif source_address_mode == MACAddressMode.SHORT:
                fcf_srcaddrmode = 2
            else:
                fcf_srcaddrmode = 3
        else:
            source_address_mode = MACAddressMode.NONE
            fcf_srcaddrmode = 0

        if destination_address_mode is not None:
            if destination_address_mode == MACAddressMode.NONE:
                fcf_destaddrmode = 0
            elif destination_address_mode == MACAddressMode.SHORT:
                fcf_destaddrmode = 2
            else:
                fcf_destaddrmode = 3
        else:
            destination_address_mode = MACAddressMode.NONE
            fcf_destaddrmode = 2

        packet = Dot15d4(
                    fcf_srcaddrmode=fcf_srcaddrmode,
                    fcf_destaddrmode=fcf_destaddrmode
        )/packet

        # Build PAN ID compression field
        packet = self._choose_pan_id_compression(packet, destination_address_mode, source_address_mode)

        sequence_number = self.database.get("macBeaconSequenceNumber")
        packet.seqnum = sequence_number
        self.database.set("macBeaconSequenceNumber", (sequence_number + 1) % 256)
        self.send('phy', packet, tag='pdu')



'''
This module provides some helpers functions and constants related to 802.15.4 physical layer.
'''
from whad.phy import OQPSKModulationScheme, PhysicalLayer, Endianness

def frequency_to_channel(frequency):
    '''
    Converts a frequency (in Hz) to the corresponding 802.15.4 channel.
    '''
    return int((((frequency / 1000000) - 2405) / 5) + 11)


def channel_to_frequency(channel):
    '''
    Converts 802.15.4 channel to frequency (in Hz).
    '''
    return 1000000 * (2405 + 5 * (channel - 11))

PHYS = {
    "802.15.4-OQPSK": PhysicalLayer(
                        modulation=OQPSKModulationScheme(),
                        datarate=250000,
                        endianness=Endianness.LITTLE,
                        frequency_range=(2405000000, 2480000000),
                        maximum_packet_size=255,
                        synchronization_word=b"\x00\x00\x00\x00\xA7",
                        frequency_to_channel_function=frequency_to_channel,
                        channel_to_frequency_function=channel_to_frequency
                    )
}



"""WHAD Protocol BLE sniffing messages abstraction layer.
"""

from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper, \
    PbFieldBool
from whad.hub.ble import BleDomain

@pb_bind(BleDomain, 'sniff_adv', 1)
class SniffAdv(PbMessageWrapper):
    """BLE advertisement sniffing message class
    """
    use_extended_adv = PbFieldBool('ble.sniff_adv.use_extended_adv')
    channel = PbFieldInt('ble.sniff_adv.channel')
    bd_address = PbFieldBytes('ble.sniff_adv.bd_address')

@pb_bind(BleDomain, 'sniff_connreq', 1)
class SniffConnReq(PbMessageWrapper):
    """BLE connection request sniffing message class
    """
    show_empty_packets = PbFieldBool('ble.sniff_connreq.show_empty_packets')
    show_advertisements = PbFieldBool('ble.sniff_connreq.show_advertisements')
    channel = PbFieldInt('ble.sniff_connreq.channel')
    bd_address = PbFieldBytes('ble.sniff_connreq.bd_address')

@pb_bind(BleDomain, 'sniff_aa', 1)
class SniffAccessAddress(PbMessageWrapper):
    """BLE connection access address sniffing message class
    """
    monitored_channels = PbFieldBytes('ble.sniff_aa.monitored_channels')

@pb_bind(BleDomain, 'sniff_conn', 1)
class SniffActiveConn(PbMessageWrapper):
    """BLE connection access address sniffing message class
    """
    access_address = PbFieldInt('ble.sniff_conn.access_address')
    crc_init = PbFieldInt('ble.sniff_conn.crc_init')
    channel_map = PbFieldBytes('ble.sniff_conn.channel_map')
    hop_interval = PbFieldInt('ble.sniff_conn.hop_interval')
    hop_increment = PbFieldInt('ble.sniff_conn.hop_increment')
    monitored_channels = PbFieldBytes('ble.sniff_conn.monitored_channels')

@pb_bind(BleDomain, 'aa_disc', 1)
class AccessAddressDiscovered(PbMessageWrapper):
    """BLE connection access address discovered notification message class
    """
    access_address = PbFieldInt('ble.aa_disc.access_address')
    rssi = PbFieldInt('ble.aa_disc.rssi')
    timestamp = PbFieldInt('ble.aa_disc.timestamp')



"""WHAD Protocol PHY modulation messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.phy.phy_pb2 import SetBPSKModulationCmd
from ..message import pb_bind, PbFieldInt, PbFieldBool, PbMessageWrapper
from . import PhyDomain

@pb_bind(PhyDomain, 'mod_ask', 1)
class SetAskMod(PbMessageWrapper):
    """PHY ASK modulation message
    """

    ook = PbFieldBool('phy.mod_ask.ook')


@pb_bind(PhyDomain, 'mod_fsk', 1)
class SetFskMod(PbMessageWrapper):
    """PHY FSK modulation message
    """

    deviation = PbFieldInt('phy.mod_fsk.deviation')

@pb_bind(PhyDomain, 'mod_gfsk', 1)
class SetGfskMod(PbMessageWrapper):
    """PHY GFSK modulation message
    """

    deviation = PbFieldInt('phy.mod_gfsk.deviation')

@pb_bind(PhyDomain, 'mod_bpsk', 1)
class SetBpskMod(PbMessageWrapper):
    """PHY BPSK modulation message
    """
    
    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.phy.mod_bpsk.CopyFrom(SetBPSKModulationCmd())

@pb_bind(PhyDomain, 'mod_qpsk', 1)
class SetQpskMod(PbMessageWrapper):
    """PHY QPSK modulation message
    """

    offset = PbFieldBool('phy.mod_qpsk.offset_qpsk')

@pb_bind(PhyDomain, 'mod_4fsk', 1)
class Set4FskMod(PbMessageWrapper):
    """PHY 4FSK modulation message
    """

    deviation = PbFieldInt('phy.mod_4fsk.deviation')

@pb_bind(PhyDomain, 'mod_msk', 1)
class SetMskMod(PbMessageWrapper):
    """PHY MSK modulation message
    """

    deviation = PbFieldInt('phy.mod_msk.deviation')

@pb_bind(PhyDomain, 'mod_lora', 1)
class SetLoRaMod(PbMessageWrapper):
    """PHY LoRa modulation message
    """

    bandwidth = PbFieldInt('phy.mod_lora.bandwidth')
    sf = PbFieldInt('phy.mod_lora.spreading_factor')
    cr = PbFieldInt('phy.mod_lora.coding_rate')
    preamble_length = PbFieldInt('phy.mod_lora.preamble_length')
    enable_crc = PbFieldBool('phy.mod_lora.enable_crc')
    explicit_mode = PbFieldBool('phy.mod_lora.explicit_mode')
    invert_iq = PbFieldBool('phy.mod_lora.invert_iq')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/ble/ble.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bwhad/protocol/ble/ble.proto\x12\x03\x62le\"J\n\x0fSetBdAddressCmd\x12\x12\n\nbd_address\x18\x01 \x01(\x0c\x12#\n\taddr_type\x18\x02 \x01(\x0e\x32\x10.ble.BleAddrType\"L\n\x0bSniffAdvCmd\x12\x18\n\x10use_extended_adv\x18\x01 \x01(\x08\x12\x0f\n\x07\x63hannel\x18\x02 \x01(\r\x12\x12\n\nbd_address\x18\x03 \x01(\x0c\"\x0b\n\tJamAdvCmd\"%\n\x12JamAdvOnChannelCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"o\n\x0fSniffConnReqCmd\x12\x1a\n\x12show_empty_packets\x18\x01 \x01(\x08\x12\x1b\n\x13show_advertisements\x18\x02 \x01(\x08\x12\x0f\n\x07\x63hannel\x18\x03 \x01(\r\x12\x12\n\nbd_address\x18\x04 \x01(\x0c\"3\n\x15SniffAccessAddressCmd\x12\x1a\n\x12monitored_channels\x18\x06 \x01(\x0c\"\x9c\x01\n\x12SniffActiveConnCmd\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\x12\x10\n\x08\x63rc_init\x18\x02 \x01(\r\x12\x13\n\x0b\x63hannel_map\x18\x03 \x01(\x0c\x12\x14\n\x0chop_interval\x18\x04 \x01(\r\x12\x15\n\rhop_increment\x18\x05 \x01(\r\x12\x1a\n\x12monitored_channels\x18\x06 \x01(\x0c\"$\n\nJamConnCmd\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\"\"\n\x0bScanModeCmd\x12\x13\n\x0b\x61\x63tive_scan\x18\x01 \x01(\x08\"5\n\nAdvModeCmd\x12\x11\n\tscan_data\x18\x01 \x01(\x0c\x12\x14\n\x0cscanrsp_data\x18\x02 \x01(\x0c\"8\n\rSetAdvDataCmd\x12\x11\n\tscan_data\x18\x01 \x01(\x0c\x12\x14\n\x0cscanrsp_data\x18\x02 \x01(\x0c\"\x10\n\x0e\x43\x65ntralModeCmd\"\x9f\x02\n\x0c\x43onnectToCmd\x12\x12\n\nbd_address\x18\x01 \x01(\x0c\x12#\n\taddr_type\x18\x02 \x01(\x0e\x32\x10.ble.BleAddrType\x12\x1b\n\x0e\x61\x63\x63\x65ss_address\x18\x03 \x01(\rH\x00\x88\x01\x01\x12\x18\n\x0b\x63hannel_map\x18\x04 \x01(\x0cH\x01\x88\x01\x01\x12\x19\n\x0chop_interval\x18\x05 \x01(\rH\x02\x88\x01\x01\x12\x1a\n\rhop_increment\x18\x06 \x01(\rH\x03\x88\x01\x01\x12\x15\n\x08\x63rc_init\x18\x07 \x01(\rH\x04\x88\x01\x01\x42\x11\n\x0f_access_addressB\x0e\n\x0c_channel_mapB\x0f\n\r_hop_intervalB\x10\n\x0e_hop_incrementB\x0b\n\t_crc_init\"\x8d\x01\n\rSendRawPDUCmd\x12$\n\tdirection\x18\x01 \x01(\x0e\x32\x11.ble.BleDirection\x12\x13\n\x0b\x63onn_handle\x18\x02 \x01(\r\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x03 \x01(\r\x12\x0b\n\x03pdu\x18\x04 \x01(\x0c\x12\x0b\n\x03\x63rc\x18\x05 \x01(\r\x12\x0f\n\x07\x65ncrypt\x18\x06 \x01(\x08\"e\n\nSendPDUCmd\x12$\n\tdirection\x18\x01 \x01(\x0e\x32\x11.ble.BleDirection\x12\x13\n\x0b\x63onn_handle\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\x12\x0f\n\x07\x65ncrypt\x18\x04 \x01(\x08\"$\n\rDisconnectCmd\x12\x13\n\x0b\x63onn_handle\x18\x01 \x01(\x05\"<\n\x11PeripheralModeCmd\x12\x11\n\tscan_data\x18\x01 \x01(\x0c\x12\x14\n\x0cscanrsp_data\x18\x02 \x01(\x0c\"\n\n\x08StartCmd\"\t\n\x07StopCmd\")\n\x0fHijackMasterCmd\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\"(\n\x0eHijackSlaveCmd\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\"\'\n\rHijackBothCmd\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\"\x80\x01\n\x10SetEncryptionCmd\x12\x13\n\x0b\x63onn_handle\x18\x01 \x01(\x05\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\x12\x0e\n\x06ll_key\x18\x03 \x01(\x0c\x12\r\n\x05ll_iv\x18\x04 \x01(\x0c\x12\x0b\n\x03key\x18\x05 \x01(\x0c\x12\x0c\n\x04rand\x18\x06 \x01(\x0c\x12\x0c\n\x04\x65\x64iv\x18\x07 \x01(\x0c\"D\n\x0eReactiveJamCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0f\n\x07pattern\x18\x02 \x01(\x0c\x12\x10\n\x08position\x18\x03 \x01(\r\"\xb5\x04\n\x12PrepareSequenceCmd\x12\x30\n\x07trigger\x18\x01 \x01(\x0b\x32\x1f.ble.PrepareSequenceCmd.Trigger\x12\n\n\x02id\x18\x02 \x01(\r\x12$\n\tdirection\x18\x03 \x01(\x0e\x32\x11.ble.BleDirection\x12\x37\n\x08sequence\x18\x04 \x03(\x0b\x32%.ble.PrepareSequenceCmd.PendingPacket\x1a\x41\n\x10ReceptionTrigger\x12\x0f\n\x07pattern\x18\x01 \x01(\x0c\x12\x0c\n\x04mask\x18\x02 \x01(\x0c\x12\x0e\n\x06offset\x18\x03 \x01(\r\x1a\x32\n\x16\x43onnectionEventTrigger\x12\x18\n\x10\x63onnection_event\x18\x01 \x01(\r\x1a\x0f\n\rManualTrigger\x1a\xd8\x01\n\x07Trigger\x12=\n\treception\x18\x01 \x01(\x0b\x32(.ble.PrepareSequenceCmd.ReceptionTriggerH\x00\x12J\n\x10\x63onnection_event\x18\x02 \x01(\x0b\x32..ble.PrepareSequenceCmd.ConnectionEventTriggerH\x00\x12\x37\n\x06manual\x18\x03 \x01(\x0b\x32%.ble.PrepareSequenceCmd.ManualTriggerH\x00\x42\t\n\x07trigger\x1a\x1f\n\rPendingPacket\x12\x0e\n\x06packet\x18\x01 \x01(\x0c\" \n\x12TriggerSequenceCmd\x12\n\n\x02id\x18\x01 \x01(\r\"\x1f\n\x11\x44\x65leteSequenceCmd\x12\n\n\x02id\x18\x01 \x01(\r\"\x17\n\tTriggered\x12\n\n\x02id\x18\x01 \x01(\r\"s\n\x17\x41\x63\x63\x65ssAddressDiscovered\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x42\x07\n\x05_rssiB\x0c\n\n_timestamp\"\x8c\x01\n\x0e\x41\x64vPduReceived\x12!\n\x08\x61\x64v_type\x18\x01 \x01(\x0e\x32\x0f.ble.BleAdvType\x12\x0c\n\x04rssi\x18\x02 \x01(\x05\x12\x12\n\nbd_address\x18\x03 \x01(\x0c\x12\x10\n\x08\x61\x64v_data\x18\x04 \x01(\x0c\x12#\n\taddr_type\x18\x05 \x01(\x0e\x32\x10.ble.BleAddrType\"\xb2\x01\n\tConnected\x12\x11\n\tinitiator\x18\x01 \x01(\x0c\x12\x12\n\nadvertiser\x18\x02 \x01(\x0c\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x03 \x01(\r\x12\x13\n\x0b\x63onn_handle\x18\x08 \x01(\r\x12\'\n\radv_addr_type\x18\t \x01(\x0e\x32\x10.ble.BleAddrType\x12(\n\x0einit_addr_type\x18\n \x01(\x0e\x32\x10.ble.BleAddrType\"3\n\x0c\x44isconnected\x12\x0e\n\x06reason\x18\x01 \x01(\r\x12\x13\n\x0b\x63onn_handle\x18\x02 \x01(\r\"z\n\x0cSynchronized\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\x12\x10\n\x08\x63rc_init\x18\x02 \x01(\r\x12\x14\n\x0chop_interval\x18\x03 \x01(\r\x12\x15\n\rhop_increment\x18\x04 \x01(\r\x12\x13\n\x0b\x63hannel_map\x18\x05 \x01(\x0c\"(\n\x0e\x44\x65synchronized\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x01 \x01(\r\"3\n\x08Hijacked\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x02 \x01(\r\"O\n\x08Injected\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x02 \x01(\r\x12\x1a\n\x12injection_attempts\x18\x03 \x01(\r\"\xda\x02\n\x0eRawPduReceived\x12$\n\tdirection\x18\x01 \x01(\x0e\x32\x11.ble.BleDirection\x12\x0f\n\x07\x63hannel\x18\x02 \x01(\r\x12\x11\n\x04rssi\x18\x03 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x04 \x01(\x04H\x01\x88\x01\x01\x12\x1f\n\x12relative_timestamp\x18\x05 \x01(\x04H\x02\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x06 \x01(\x08H\x03\x88\x01\x01\x12\x16\n\x0e\x61\x63\x63\x65ss_address\x18\x07 \x01(\r\x12\x0b\n\x03pdu\x18\x08 \x01(\x0c\x12\x0b\n\x03\x63rc\x18\t \x01(\r\x12\x13\n\x0b\x63onn_handle\x18\n \x01(\r\x12\x11\n\tprocessed\x18\x0b \x01(\x08\x12\x11\n\tdecrypted\x18\x0c \x01(\x08\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x15\n\x13_relative_timestampB\x0f\n\r_crc_validity\"{\n\x0bPduReceived\x12$\n\tdirection\x18\x01 \x01(\x0e\x32\x11.ble.BleDirection\x12\x0b\n\x03pdu\x18\x02 \x01(\x0c\x12\x13\n\x0b\x63onn_handle\x18\x03 \x01(\r\x12\x11\n\tprocessed\x18\x04 \x01(\x08\x12\x11\n\tdecrypted\x18\x05 \x01(\x08\"\xbc\x0c\n\x07Message\x12+\n\x0bset_bd_addr\x18\x01 \x01(\x0b\x32\x14.ble.SetBdAddressCmdH\x00\x12%\n\tsniff_adv\x18\x02 \x01(\x0b\x32\x10.ble.SniffAdvCmdH\x00\x12!\n\x07jam_adv\x18\x03 \x01(\x0b\x32\x0e.ble.JamAdvCmdH\x00\x12/\n\x0cjam_adv_chan\x18\x04 \x01(\x0b\x32\x17.ble.JamAdvOnChannelCmdH\x00\x12-\n\rsniff_connreq\x18\x05 \x01(\x0b\x32\x14.ble.SniffConnReqCmdH\x00\x12.\n\x08sniff_aa\x18\x06 \x01(\x0b\x32\x1a.ble.SniffAccessAddressCmdH\x00\x12-\n\nsniff_conn\x18\x07 \x01(\x0b\x32\x17.ble.SniffActiveConnCmdH\x00\x12#\n\x08jam_conn\x18\x08 \x01(\x0b\x32\x0f.ble.JamConnCmdH\x00\x12%\n\tscan_mode\x18\t \x01(\x0b\x32\x10.ble.ScanModeCmdH\x00\x12#\n\x08\x61\x64v_mode\x18\n \x01(\x0b\x32\x0f.ble.AdvModeCmdH\x00\x12*\n\x0cset_adv_data\x18\x0b \x01(\x0b\x32\x12.ble.SetAdvDataCmdH\x00\x12+\n\x0c\x63\x65ntral_mode\x18\x0c \x01(\x0b\x32\x13.ble.CentralModeCmdH\x00\x12$\n\x07\x63onnect\x18\r \x01(\x0b\x32\x11.ble.ConnectToCmdH\x00\x12*\n\x0csend_raw_pdu\x18\x0e \x01(\x0b\x32\x12.ble.SendRawPDUCmdH\x00\x12#\n\x08send_pdu\x18\x0f \x01(\x0b\x32\x0f.ble.SendPDUCmdH\x00\x12(\n\ndisconnect\x18\x10 \x01(\x0b\x32\x12.ble.DisconnectCmdH\x00\x12-\n\x0bperiph_mode\x18\x11 \x01(\x0b\x32\x16.ble.PeripheralModeCmdH\x00\x12\x1e\n\x05start\x18\x12 \x01(\x0b\x32\r.ble.StartCmdH\x00\x12\x1c\n\x04stop\x18\x13 \x01(\x0b\x32\x0c.ble.StopCmdH\x00\x12-\n\rhijack_master\x18\x14 \x01(\x0b\x32\x14.ble.HijackMasterCmdH\x00\x12+\n\x0chijack_slave\x18\x15 \x01(\x0b\x32\x13.ble.HijackSlaveCmdH\x00\x12)\n\x0bhijack_both\x18\x16 \x01(\x0b\x32\x12.ble.HijackBothCmdH\x00\x12+\n\nencryption\x18! \x01(\x0b\x32\x15.ble.SetEncryptionCmdH\x00\x12+\n\x0creactive_jam\x18\" \x01(\x0b\x32\x13.ble.ReactiveJamCmdH\x00\x12*\n\x07prepare\x18# \x01(\x0b\x32\x17.ble.PrepareSequenceCmdH\x00\x12*\n\x07trigger\x18$ \x01(\x0b\x32\x17.ble.TriggerSequenceCmdH\x00\x12,\n\ndelete_seq\x18& \x01(\x0b\x32\x16.ble.DeleteSequenceCmdH\x00\x12/\n\x07\x61\x61_disc\x18\x17 \x01(\x0b\x32\x1c.ble.AccessAddressDiscoveredH\x00\x12&\n\x07\x61\x64v_pdu\x18\x18 \x01(\x0b\x32\x13.ble.AdvPduReceivedH\x00\x12#\n\tconnected\x18\x19 \x01(\x0b\x32\x0e.ble.ConnectedH\x00\x12)\n\x0c\x64isconnected\x18\x1a \x01(\x0b\x32\x11.ble.DisconnectedH\x00\x12)\n\x0csynchronized\x18\x1b \x01(\x0b\x32\x11.ble.SynchronizedH\x00\x12!\n\x08hijacked\x18\x1c \x01(\x0b\x32\r.ble.HijackedH\x00\x12\x1f\n\x03pdu\x18\x1d \x01(\x0b\x32\x10.ble.PduReceivedH\x00\x12&\n\x07raw_pdu\x18\x1e \x01(\x0b\x32\x13.ble.RawPduReceivedH\x00\x12!\n\x08injected\x18\x1f \x01(\x0b\x32\r.ble.InjectedH\x00\x12-\n\x0e\x64\x65synchronized\x18  \x01(\x0b\x32\x13.ble.DesynchronizedH\x00\x12#\n\ttriggered\x18% \x01(\x0b\x32\x0e.ble.TriggeredH\x00\x42\x05\n\x03msg*\xcf\x03\n\nBleCommand\x12\x10\n\x0cSetBdAddress\x10\x00\x12\x0c\n\x08SniffAdv\x10\x01\x12\n\n\x06JamAdv\x10\x02\x12\x13\n\x0fJamAdvOnChannel\x10\x03\x12\x0f\n\x0bReactiveJam\x10\x04\x12\x10\n\x0cSniffConnReq\x10\x05\x12\x16\n\x12SniffAccessAddress\x10\x06\x12\x13\n\x0fSniffActiveConn\x10\x07\x12\x0b\n\x07JamConn\x10\x08\x12\x0c\n\x08ScanMode\x10\t\x12\x0b\n\x07\x41\x64vMode\x10\n\x12\x0e\n\nSetAdvData\x10\x0b\x12\x0f\n\x0b\x43\x65ntralMode\x10\x0c\x12\r\n\tConnectTo\x10\r\x12\x0e\n\nSendRawPDU\x10\x0e\x12\x0b\n\x07SendPDU\x10\x0f\x12\x0e\n\nDisconnect\x10\x10\x12\x12\n\x0ePeripheralMode\x10\x11\x12\t\n\x05Start\x10\x12\x12\x08\n\x04Stop\x10\x13\x12\x11\n\rSetEncryption\x10\x14\x12\x10\n\x0cHijackMaster\x10\x15\x12\x0f\n\x0bHijackSlave\x10\x16\x12\x0e\n\nHijackBoth\x10\x17\x12\x13\n\x0fPrepareSequence\x10\x18\x12\x13\n\x0fTriggerSequence\x10\x19\x12\x12\n\x0e\x44\x65leteSequence\x10\x1a*w\n\nBleAdvType\x12\x0f\n\x0b\x41\x44V_UNKNOWN\x10\x00\x12\x0b\n\x07\x41\x44V_IND\x10\x01\x12\x12\n\x0e\x41\x44V_DIRECT_IND\x10\x02\x12\x13\n\x0f\x41\x44V_NONCONN_IND\x10\x03\x12\x10\n\x0c\x41\x44V_SCAN_IND\x10\x04\x12\x10\n\x0c\x41\x44V_SCAN_RSP\x10\x05*v\n\x0c\x42leDirection\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x13\n\x0fMASTER_TO_SLAVE\x10\x01\x12\x13\n\x0fSLAVE_TO_MASTER\x10\x02\x12\x16\n\x12INJECTION_TO_SLAVE\x10\x03\x12\x17\n\x13INJECTION_TO_MASTER\x10\x04*%\n\x0b\x42leAddrType\x12\n\n\x06PUBLIC\x10\x00\x12\n\n\x06RANDOM\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.ble.ble_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _BLECOMMAND._serialized_start=5287
  _BLECOMMAND._serialized_end=5750
  _BLEADVTYPE._serialized_start=5752
  _BLEADVTYPE._serialized_end=5871
  _BLEDIRECTION._serialized_start=5873
  _BLEDIRECTION._serialized_end=5991
  _BLEADDRTYPE._serialized_start=5993
  _BLEADDRTYPE._serialized_end=6030
  _SETBDADDRESSCMD._serialized_start=36
  _SETBDADDRESSCMD._serialized_end=110
  _SNIFFADVCMD._serialized_start=112
  _SNIFFADVCMD._serialized_end=188
  _JAMADVCMD._serialized_start=190
  _JAMADVCMD._serialized_end=201
  _JAMADVONCHANNELCMD._serialized_start=203
  _JAMADVONCHANNELCMD._serialized_end=240
  _SNIFFCONNREQCMD._serialized_start=242
  _SNIFFCONNREQCMD._serialized_end=353
  _SNIFFACCESSADDRESSCMD._serialized_start=355
  _SNIFFACCESSADDRESSCMD._serialized_end=406
  _SNIFFACTIVECONNCMD._serialized_start=409
  _SNIFFACTIVECONNCMD._serialized_end=565
  _JAMCONNCMD._serialized_start=567
  _JAMCONNCMD._serialized_end=603
  _SCANMODECMD._serialized_start=605
  _SCANMODECMD._serialized_end=639
  _ADVMODECMD._serialized_start=641
  _ADVMODECMD._serialized_end=694
  _SETADVDATACMD._serialized_start=696
  _SETADVDATACMD._serialized_end=752
  _CENTRALMODECMD._serialized_start=754
  _CENTRALMODECMD._serialized_end=770
  _CONNECTTOCMD._serialized_start=773
  _CONNECTTOCMD._serialized_end=1060
  _SENDRAWPDUCMD._serialized_start=1063
  _SENDRAWPDUCMD._serialized_end=1204
  _SENDPDUCMD._serialized_start=1206
  _SENDPDUCMD._serialized_end=1307
  _DISCONNECTCMD._serialized_start=1309
  _DISCONNECTCMD._serialized_end=1345
  _PERIPHERALMODECMD._serialized_start=1347
  _PERIPHERALMODECMD._serialized_end=1407
  _STARTCMD._serialized_start=1409
  _STARTCMD._serialized_end=1419
  _STOPCMD._serialized_start=1421
  _STOPCMD._serialized_end=1430
  _HIJACKMASTERCMD._serialized_start=1432
  _HIJACKMASTERCMD._serialized_end=1473
  _HIJACKSLAVECMD._serialized_start=1475
  _HIJACKSLAVECMD._serialized_end=1515
  _HIJACKBOTHCMD._serialized_start=1517
  _HIJACKBOTHCMD._serialized_end=1556
  _SETENCRYPTIONCMD._serialized_start=1559
  _SETENCRYPTIONCMD._serialized_end=1687
  _REACTIVEJAMCMD._serialized_start=1689
  _REACTIVEJAMCMD._serialized_end=1757
  _PREPARESEQUENCECMD._serialized_start=1760
  _PREPARESEQUENCECMD._serialized_end=2325
  _PREPARESEQUENCECMD_RECEPTIONTRIGGER._serialized_start=1939
  _PREPARESEQUENCECMD_RECEPTIONTRIGGER._serialized_end=2004
  _PREPARESEQUENCECMD_CONNECTIONEVENTTRIGGER._serialized_start=2006
  _PREPARESEQUENCECMD_CONNECTIONEVENTTRIGGER._serialized_end=2056
  _PREPARESEQUENCECMD_MANUALTRIGGER._serialized_start=2058
  _PREPARESEQUENCECMD_MANUALTRIGGER._serialized_end=2073
  _PREPARESEQUENCECMD_TRIGGER._serialized_start=2076
  _PREPARESEQUENCECMD_TRIGGER._serialized_end=2292
  _PREPARESEQUENCECMD_PENDINGPACKET._serialized_start=2294
  _PREPARESEQUENCECMD_PENDINGPACKET._serialized_end=2325
  _TRIGGERSEQUENCECMD._serialized_start=2327
  _TRIGGERSEQUENCECMD._serialized_end=2359
  _DELETESEQUENCECMD._serialized_start=2361
  _DELETESEQUENCECMD._serialized_end=2392
  _TRIGGERED._serialized_start=2394
  _TRIGGERED._serialized_end=2417
  _ACCESSADDRESSDISCOVERED._serialized_start=2419
  _ACCESSADDRESSDISCOVERED._serialized_end=2534
  _ADVPDURECEIVED._serialized_start=2537
  _ADVPDURECEIVED._serialized_end=2677
  _CONNECTED._serialized_start=2680
  _CONNECTED._serialized_end=2858
  _DISCONNECTED._serialized_start=2860
  _DISCONNECTED._serialized_end=2911
  _SYNCHRONIZED._serialized_start=2913
  _SYNCHRONIZED._serialized_end=3035
  _DESYNCHRONIZED._serialized_start=3037
  _DESYNCHRONIZED._serialized_end=3077
  _HIJACKED._serialized_start=3079
  _HIJACKED._serialized_end=3130
  _INJECTED._serialized_start=3132
  _INJECTED._serialized_end=3211
  _RAWPDURECEIVED._serialized_start=3214
  _RAWPDURECEIVED._serialized_end=3560
  _PDURECEIVED._serialized_start=3562
  _PDURECEIVED._serialized_end=3685
  _MESSAGE._serialized_start=3688
  _MESSAGE._serialized_end=5284
# @@protoc_insertion_point(module_scope)



"""This module provides a connector to use LoRaWAN capable hardware.
"""
import logging

from time import sleep
from queue import Queue, Empty
from struct import unpack
from Cryptodome.Cipher import AES
from Cryptodome.Hash import CMAC
from scapy.contrib.loraphy2wan import PHYPayload, Join_Request, Join_Accept

from whad.device import WhadDevice
from whad.phy.connector.lora import LoRa

from ..channel import EU868, ChannelModParams, ChannelPlan
from ..exceptions import NotStartedException

logger = logging.getLogger(__name__)

def eui_to_bytes(eui) -> bytes:
    """Convert EUI to bytes.

    :param eui: LoRaWAN EUI in the form of comma-separated hex values
    :type eui: str
    """
    values = eui.lower().split(':')[::-1]
    assert len(values) == 8
    return bytes([int(v, 16) for v in values])

def compute_mic(appkey, buffer) -> bytes:
    """Compute a message MIC given an APP key.
    
    :param appkey: APP key
    :type appkey: bytes
    :param buffer: Message
    :type buffer: bytes
    :return: Computed MIC (4 bytes)
    :rtype: bytes
    """
    c = CMAC.new(appkey, ciphermod=AES)
    c.update(buffer)
    return c.digest()[:4]


class LoRaWAN(LoRa):
    '''Basic LoRaWAN connector to use with a LoRa compatible hardware.

    This connector provides the basic features to send and receive LoRaWAN packets
    with a specific channel plan.
    '''

    def __init__(self, device : WhadDevice = None, channel_plan : ChannelPlan = EU868):
        """Initialize this LoRaWAN connector.

        :param device: Compatible device to use.
        :type device: :class:`whad.device.WhadDevice`
        :param channel_plan: Channel plan to use
        :type channel_plan: :class:`whad.lorawan.channel.ChannelPlan`
        """
        super().__init__(device)

        # Configure channel plan
        logger.debug("Channel plan set to %s", channel_plan.__name__)
        self.__channel_plan = channel_plan()
        self.__pkt_queue = Queue()
        self.__cr = 45

        # Keep track of current status
        self.__started = False

        # Default mode is uplink
        self.uplink()

    def reconfigure(self, channel: ChannelModParams, crc: bool = True, invert_iq : bool = False):
        '''Reconfigure hardware with provided parameters.

        :param channel: Target channel modulation parameters.
        :type channel: :class:`whad.lorawan.channel.ChannelModParams`
        :param crc: Enable CRC if set to `True`, disabled if set to `False`
        :type crc: bool, optional
        :param invert_iq: Invert IQ if set to `True`
        :type invert_iq: bool, optional
        '''
        logger.debug("Reconfiguring hardware for channel %s", channel)

        # Reconfigure hardware (stop if necessary, then restart if we were already started)
        must_restart = self.__started
        if self.__started:
            logger.debug('Hardware was running, stopping ...')
            self.stop()
            logger.debug('Hardware stopped')

        # Set modulation parameters
        logger.debug('Reconfiguring core modulation parameters')
        self.sf = channel.spreading_factor
        self.bw = channel.bandwidth
        self.preamble_length = 8
        self.invert_iq = invert_iq
        logger.debug("Enabling CRC: %s", crc)
        self.enable_crc(crc)
        self.enable_explicit_mode(True)
        logger.debug("Setting Freq: %d", channel.frequency)
        self.set_frequency(channel.frequency)
        self.syncword = LoRa.SYNCWORD_LORAWAN

        # Restart hardware if required
        if must_restart:
            logger.debug('Resuming RX ...')
            self.start()

    def uplink(self):
        '''Configure hardware to transmit on a random uplink channel.

        A random uplink channel is picked from the ones defined in the channel
        plan, and hardware is reconfigured to listen and send on the corresponding
        frequency with the associated modulation parameters.
        '''
        self.__current_channel = self.__channel_plan.pick_channel()
        self.reconfigure(self.__current_channel)
        logger.debug("TX channel: %s", self.__current_channel)

    def rx1(self):
        '''Configure hardware to listen on RX1.
        
        RX1 channel is chosen depending on the channel plan.
        '''
        # Retrieve RX1 channel modulation parameters from channel plan
        rx1_channel = self.__channel_plan.get_rx1(self.__current_channel.number)
        logger.debug("RX1 channel: %s", rx1_channel)

        # Change hardware configuration only if needed
        if rx1_channel != self.__current_channel or self.crc_enabled:
            self.reconfigure(rx1_channel, crc=False, invert_iq=True)

    def rx2(self):
        '''Configure hardware to listen on RX2.

        RX2 channel is usually a single channel with more reliable modulation
        parameters used as a backup channel for downlink communication.
        '''
        rx2_channel = self.__channel_plan.get_rx2()
        logger.debug("RX2 channel: %s", rx2_channel)

        # Change hardware configuration only if needed
        if rx2_channel != self.__current_channel or self.crc_enabled:
            self.reconfigure(rx2_channel, crc=False, invert_iq=True)


    def start(self, coding_rate: int=45):
        """Start the LoRaWAN adapter into receive mode by default.
        """
        # Start listening
        logger.debug('Starting hardware (RX mode)')
        super().start()
        self.__started = True


    def stop(self):
        """Stop the LoRaWAN adapter.
        """
        logger.debug('Stopping hardware')
        super().stop()
        self.__started = False


    def send(self, packet, timestamp: float = None):
        '''Send a LoRaWAN frame with current LoRa modulation parameters.

        :param packet: Packet to send
        :type packet: :class:`whad.scapy.layers.lorawan.PHYPayload`
        :param timestamp: If provided, will send the packet at the given timestamp
        :type timestamp: float, optional
        '''
        # Make sure hardware has been started
        if not self.__started:
            raise NotStartedException

        # Send LoRaWAN frame
        if timestamp is not None:
            logger.debug("Programming packet %s at %f", bytes(packet).hex(), timestamp)
            pkt_id = super().schedule_send(packet, timestamp)
            logger.debug("packet id is %d", pkt_id)
            return pkt_id
        else:
            logger.debug("Sending packet %s", bytes(packet).hex())
            super().send(packet)


    def on_packet(self, packet : PHYPayload):
        """Callback method for incoming packet processing

        :param packet: Received packet
        :type packet: :class:`whad.scapy.layers.lorawan.PHYPayload`
        """
        logger.debug(f'Received LoRaWAN payload: {bytes(packet).hex()}')

        # Add packet to our packet queue
        pkt = PHYPayload(bytes(packet))
        pkt.metadata = packet.metadata
        self.__pkt_queue.put(pkt)


    def wait_packet(self, timeout : float = None):
        """Wait for a LoRaWAN packet.

        If timeout is set, wait for the given time and return None if no
        packet has been received. If timeout is not provided, this method
        will block until a valid packet is received.

        :param timeout: Timeout in seconds
        :type timeout: float, optional
        """
        try:
            logger.debug('Waiting for incoming packet ...')
            packet = self.__pkt_queue.get(block=True, timeout=timeout)           
            logger.debug('Packet received !')
            return packet
        except Empty:
            return None

    def __process_join_accept(self, app_key, packet):
        '''Process an encrypted JoinAccept
        '''
        # Decrypt and verify JoinAccept
        if len(packet.Join_Accept_Encrypted) > 0:
            ja_enc = packet.Join_Accept_Encrypted
            
            # Decrypt JoinAccept
            c = AES.new(app_key, mode=AES.MODE_ECB)
            ja_dec = c.encrypt(ja_enc)
            ja_dec, mic = ja_dec[:-4], ja_dec[-4:]
            resp = Join_Accept(ja_dec)

            logger.debug("Decrypted JoinAccept: %s", ja_dec.hex())
            logger.debug("JoinAccept MIC: %s", mic.hex())

            # Check MIC
            buf = b'\x20' + ja_dec
            exp_mic = compute_mic(app_key, buf)
            if exp_mic == mic:
                logger.debug('MIC is OK')
                # Return JoinAccept if MIC is OK
                return resp
            else:
                logger.debug("MIC does not match (expected: %s)", exp_mic)
                return None

    def join(self, app_key : bytes, app_eui : str, dev_eui : str, dev_nonce : int):
        '''Perform an OTAA join procedure

        See section 6.2.1 from LoRaWAN Specifications version 1.1

        :param app_key: Application key
        :type app_key: bytes
        :param app_eui: Application EUI
        :type app_eui: str
        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_nonce: Device nonce
        :type dev_nonce: str
        '''
        logger.debug(
            "Building a join request for APPEUI %s, DEVEUI %s",
            app_eui,
            dev_eui,
        )
        logger.debug("Using DevNonce=0x%04x", dev_nonce)

        join_req = Join_Request()
        join_req.AppEUI = eui_to_bytes(app_eui)
        join_req.DevEUI = eui_to_bytes(dev_eui)
        join_req.DevNonce = dev_nonce

        logging.debug('Generating PHY payload from JoinRequest ...')
        phy_jr = PHYPayload()
        phy_jr.MType = 0x00 # join request
        phy_jr.Join_Request_Field = join_req
        mic = compute_mic(app_key, bytes(phy_jr)[:-4])
        phy_jr.MIC = unpack('>I', mic)[0]
        logging.debug("PHY[JoinRequest]: %s", bytes(phy_jr).hex())
        logging.debug(" - MIC: %s", mic.hex())

        # Send join request
        self.send(bytes(phy_jr))

        # Wait for join request to be sent
        sleep(.3)

        # Switch to RX1
        logger.debug('Opening RX1 window...')
        self.rx1()

        # Wait for a join accept
        ja = self.wait_packet(5.5)
        if ja is not None:
            # Decrypt and verify JoinAccept
            result = self.__process_join_accept(app_key, ja)
            if result is not None:
                logger.debug('Received a JoinAccept on RX1')
                return result
                    
        # Switch to RX2
        logger.debug('Opening RX2 window...')
        self.rx2()

        # Wait for a JoinAccept (again)
        ja = self.wait_packet(5.5)
        if ja is not None:
            # Decrypt and verify JoinAccept
            result = self.__process_join_accept(ja)
            if result is not None:
                logger.debug('Received a JoinAccept on RX2')
                return result
            
        # Join failed
        logger.debug('No JoinAccept received, join procedure FAILED')

        # Switch back to uplink
        self.uplink()

        return None
        


"""User-defined transform tool
"""
import logging
from time import sleep

from scapy.config import conf
from scapy.themes import BrightTheme

from whad.cli.app import CommandLineApp, run_app
from whad.device.unix import UnixSocketServer, UnixConnector
from whad.device import Bridge
from whad.hub import ProtocolHub

logger = logging.getLogger(__name__)

class UserTransformPipe(Bridge):
    """User-defined transform pipe
    """
    def __init__(self, input_connector, output_connector, on_rx_packet_cb, on_tx_packet_cb):
        super().__init__(input_connector, output_connector)
        self.on_rx_packet = on_rx_packet_cb
        self.on_tx_packet = on_tx_packet_cb


    def on_outbound(self, message):
        """Process outbound messages.

        Outbound packets are packets coming from our input connector,that need to be
        forwarded as packets to the next tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            if pkt is not None:
                pkt = self.on_rx_packet(pkt)
                if pkt is not None:
                    msg = message.from_packet(pkt)
                    super().on_outbound(msg)
        else:
            logger.debug(
                '[user-transform][input-pipe] forward default outbound message %s',
                message
            )
            # Forward other messages
            super().on_outbound(message)


    def on_inbound(self, message):
        """Process inbound messages.

        Inbound packets are packets coming from our output connector,that need to be
        forwarded as packets to the previous tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            if pkt is not None:
                pkt = self.on_tx_packet(pkt)
                if pkt is not None:
                    msg = message.from_packet(pkt)
                    super().on_inbound(msg)
        else:
            logger.debug(
                '[user-transform][input-pipe] forward default inbound message %s',
                message
            )
            # Forward other messages
            super().on_inbound(message)


class UserTransformApp(CommandLineApp):
    """User-defined transform command-line application

    This class defines a configurable transform application that calls user-defined
    callbacks when packets flows in and out.
    """

    def __init__(self, inbound_cb = None, outbound_cb = None):
        """Application uses an interface and has commands.
        """
        # Save callbacks
        self.__inbound_cb = inbound_cb
        self.__outbound_cb = outbound_cb

        # Initialize user-defined transform app
        super().__init__(
            description='User transform application',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_WHAD
        )

    def on_rx_packet(self, pkt):
        """Process inbound packet
        """
        logger.debug("[user-transform][on_rx_packet] inbound packet received")
        if self.__inbound_cb is not None:
            pkt = self.__inbound_cb(pkt)
        return pkt


    def on_tx_packet(self, pkt):
        """Process outbound packet
        """
        logger.debug("[user-transform][on_tx_packet] outbound packet received")
        if self.__outbound_cb is not None:
            pkt = self.__outbound_cb(pkt)
        return pkt

    def run(self):
        """Application's main task
        """
        # Launch pre-run tasks
        self.pre_run()

        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                parameters = self.args.__dict__
                connector = UnixConnector(interface)

                connector.domain = self.args.domain
                hub = ProtocolHub(2)
                connector.format = hub.get(self.args.domain).format

                if self.is_stdout_piped():
                    unix_server = UnixConnector(UnixSocketServer(parameters=parameters))

                    while not unix_server.device.opened:
                        if unix_server.device.timedout:
                            return
                        else:
                            sleep(0.1)
                    # Create our packet bridge
                    logger.debug("[user-transform] Starting our output pipe")
                    _ = UserTransformPipe(
                        connector,
                        unix_server,
                        self.on_rx_packet,
                        self.on_tx_packet
                    )

                else:
                    connector.on_packet = self.on_rx_packet

                # Once configured, unlock connector to enable packet processing
                connector.unlock()

                # Keep running while interface is active
                while interface.opened:
                    sleep(.1)
            else:
                exit(1)
        except KeyboardInterrupt:
            # Launch post-run tasks
            self.post_run()

def user_transform(inbound_cb, outbound_cb):
    """User-defined transform application wrapper
    """
    app = UserTransformApp(inbound_cb, outbound_cb)
    run_app(app)



class MissingNetworkSecurityHeader(Exception):
    """
    Exception raised when no network security header can be found during encryption / decryption.
    """
    def __init__(self):
        super().__init__()



"""
This module provides some constants used by WHAD to communicate with the RZUSBStick.
"""
from enum import IntEnum

# USB device reset
class RZUSBStickOSPrimitive(IntEnum):
    """OS Primitive enumeration.
    """
    USBDEVFS_RESET = ord('U') << (4 * 2) | 20

# RZUSBStick Internal states
class RZUSBStickInternalStates(IntEnum):
    """Internal states.
    """
    NONE                    = 0
    SNIFFING                = 1
    TRANSMITTING            = 2

# USB identifiers
class RZUSBStickId(IntEnum):
    """USB vendor and product ID.
    """
    RZUSBSTICK_ID_VENDOR    = 0x03eb
    RZUSBSTICK_ID_PRODUCT   = 0x210a

# USB Endpoints
class RZUSBStickEndPoints(IntEnum):
    """USB endpoints.
    """
    RZ_COMMAND_ENDPOINT     = 0x02
    RZ_RESPONSE_ENDPOINT    = 0x84
    RZ_PACKET_ENDPOINT 	    = 0x81

# USB Commands
class RZUSBStickCommands(IntEnum):
    """RZ USB commands.
    """
    RZ_SET_MODE             = 0x07
    RZ_SET_CHANNEL          = 0x08
    RZ_OPEN_STREAM          = 0x09
    RZ_CLOSE_STREAM	        = 0x0A
    RZ_INJECT_FRAME         = 0x0D
    RZ_JAMMER_ON            = 0x0E
    RZ_JAMMER_OFF           = 0x0F

# USB Responses
class RZUSBStickResponses(IntEnum):
    """USB responses.
    """
    RZ_RESP_SUCCESS         = 0x80
    RZ_AIRCAPTURE_DATA      = 0x50

# RZ Modes
class RZUSBStickModes(IntEnum):
    """RZ USB stick mode.
    """
    RZ_MODE_AIRCAPTURE      = 0x00
    RZ_MODE_NONE            = 0x04



"""
WHAD Enhanced ShockBurst injection connector.

This module provides the `Injector` class for ESB-enable devices.
This class is used by `winject` to perform packet/PDU injection into an
existing ESB connection, or simply send ESB packets in the air.
"""
from scapy.packet import Packet

from .ptx import PTX
from ..injecting import InjectionConfiguration

class Injector(PTX):
    """
    Enhanced ShockBurst Injector interface.
    """
    def __init__(self, device):
        super().__init__(device)
        self._synced = False
        self._configuration = InjectionConfiguration()
        self._injecting = False

    @property
    def configuration(self) -> InjectionConfiguration:
        """Retrieve this injector configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: InjectionConfiguration):
        """Set the injector configuration.
        """
        self.stop()
        self.__configuration = new_configuration
        if self.__configuration.address is not None:
            self.address = self.__configuration.address
        if self.__configuration.channel is not None:
            self.channel = self.__configuration.channel

        self.start()

    def on_pdu(self, packet: Packet):
        """Callback method to handle incoming packet

        :param packet: Incoming packet
        :type packet: Packet
        """
        # If we are injecting, do not process incoming packets.
        if self._injecting:
            return None

        # Let ESB class process the incoming packet.
        return super().on_pdu(packet)

    def inject(self, packet: Packet):
        """Perform packet injection.

        :param packet: Packet to inject
        :type packet: Packet
        """
        if hasattr(packet, "address") and packet.address != self.address:
            self.stop()
            self.address = packet.address
            self.start()

        if self.__configuration.synchronize:
            # Synchronize if not already sync'ed
            if not self._synced:
                self.synchronize()
                self._injecting = True
                while not self.send(packet, channel=self.channel, address=self.address):
                    self._injecting = False
                    self.synchronize()
                self._synced = True
                return True

            # If already synchronized, just send packet.
            return self.send(packet, channel=self.channel, address=self.address)

        return self.send(packet, channel=self.__configuration.channel, address=self.address)



from whad.common.analyzer import TrafficAnalyzer
from whad.scapy.layers.rf4ce import RC_COMMAND_CODES, RF4CE_Vendor_MSO_Audio_Start_Request, \
    RF4CE_Vendor_ZRC_User_Control_Pressed, RF4CE_Vendor_MSO_User_Control_Pressed, \
    RF4CE_Vendor_MSO_Audio_Data_Notify, RF4CE_Vendor_MSO_Audio_Stop_Request, \
    RF4CE_Vendor_MSO_Audio
from whad.rf4ce.utils.adpcm import ADPCM
from whad.rf4ce.crypto import RF4CEKeyDerivation
import os, tempfile

class RF4CEAudio(TrafficAnalyzer):
    def __init__(self):
        super().__init__()

    @property
    def output(self):
        return {
            "raw_audio" :self.raw_audio
        }

    def reset(self):
        super().reset()
        self.audio_filename = os.path.join(tempfile.mkdtemp()+ '.wav')
        self.adpcm = ADPCM(live_play=False, output_filename=self.audio_filename)
        self.raw_audio = None

    def process_packet(self, packet):
        if RF4CE_Vendor_MSO_Audio_Start_Request in packet:
            self.trigger()
            self.adpcm.process_packet(packet)
            self.mark_packet(packet)

        elif RF4CE_Vendor_MSO_Audio_Data_Notify in packet:
            self.adpcm.process_packet(packet)
            self.mark_packet(packet)

        elif RF4CE_Vendor_MSO_Audio_Stop_Request in packet or (RF4CE_Vendor_MSO_Audio in packet and packet.audio_cmd_id == 2):
            self.adpcm.process_packet(packet)
            self.mark_packet(packet)
            with open(self.audio_filename, "rb") as f:
                self.raw_audio = f.read()
            os.unlink(self.audio_filename)
            self.complete()

class RF4CEKeystroke(TrafficAnalyzer):

    @property
    def output(self):
        return {
            "key" :self.key
        }

    def process_packet(self, packet):
        code = None
        if RF4CE_Vendor_ZRC_User_Control_Pressed in packet or RF4CE_Vendor_MSO_User_Control_Pressed in packet:
            self.trigger()
            self.mark_packet(packet)
            code = packet.code


        if code is not None:
            try:
                key = RC_COMMAND_CODES[code]

                if key != self.key:
                    self.key = key
                    if len(self.key) > 1:
                        self.key = " [{}] ".format(self.key)
                    self.complete()
            except KeyError:
                self.reset()

    def reset(self):
        super().reset()
        self.key = None

analyzers = {
    "key_cracking" : RF4CEKeyDerivation,
    "audio" : RF4CEAudio,
    "keystroke" : RF4CEKeystroke
}



"""Bluetooth Low Energy utility for WHAD

This utility is exposed in WHAD as `wble-central` and provides the following
features:

- BLE scanning
- GATT client
- Device emulation

$ wble-central scan    -> find the first usable BLE-capable device
$ wble-central -i uart scan
$ wble-central connect <bd address> <random/public>
 -> gatttool like, spawns a shell
$ wble-central -> spawns a shell
$ wble-central emulate myfile.json
$ wble-central sniff <bd address> -> capture connections to this device and save to a pcap

"""
import logging
from whad.cli.app import CommandLineApp, run_app

from whad.ble.utils.clues import CluesDb

from .shell import BleCentralShell
from .commands import *

logger = logging.getLogger(__name__)

class BleCentralApp(CommandLineApp):
    """BLE Central-role application.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD Bluetooth Low Energy central utility',
            interface=True,
            commands=True,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_STANDARD
        )

        self.add_argument(
            "--bdaddr",
            "-b",
            dest="bdaddr",
            help="Specify target BD address"
        )

        self.add_argument(
            "--random",
            "-r",
            action="store_true",
            dest="random",
            default=False,
            help="If set, target device has a random BD address (default: public)"
        )

        self.add_argument(
            "--spoof-public",
            "-s",
            dest="bdaddr_pub_src",
            help="Specify a public BD address to spoof"
        )

        self.add_argument(
            "--spoof-random",
            dest="bdaddr_rand_src",
            help="Specify a random BD address to spoof"
        )

        self.add_argument(
            "--file",
            "-f",
            dest="script",
            help="Specify a script to run"
        )

        self.add_argument(
            "--profile",
            "-p",
            dest="profile",
            help="Use a saved device profile"
        )


    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        # Launch pre-run tasks
        logger.debug("Executing pre-run hook")
        self.pre_run()

        # Pre-load CluesDb
        if not CluesDb.load_data():
            self.error("CLUES database could not be loaded, did you use `--recurse-submodules` when cloning the repository ?")

        logger.debug("Executing main code")
        if self.args.script is not None:
            logger.debug("Checking is input interface is piped")
            if self.is_piped_interface():
                # Make sure we have all the required parameters
                failed = False
                for param in ['initiator_bdaddr', 'initiator_addrtype', 'target_bdaddr', 'target_addrtype', 'conn_handle']:
                    if not hasattr(self.args, param):
                        self.error('Source interface does not provide a BLE connection')
                        failed = True
                        break

                if not failed:
                    # Create central device
                    central, _ = create_central(self, piped=True)

                    if central is not None:
                        myshell = BleCentralShell(
                            self.input_interface,
                            connector=central,
                            bd_address=self.args.target_bdaddr
                        )
                        myshell.run_script(self.args.script)
                    else:
                        self.error("Failed to open piped interface.")

            # We need to have an interface specified
            elif self.interface is not None:
                # Launch an interactive shell (well, driven by our script)
                logger.debug("Launching interactive shell with interface %s", self.interface)
                myshell = BleCentralShell(self.interface)
                myshell.run_script(self.args.script)
        else:
            super().run(pre=False, post=False)

        # Launch post-run tasks
        logger.debug("Executing post-run hook")
        self.post_run()

def ble_central_main():
    """BLE Central application main routine.
    """
    # Run our central app
    app = BleCentralApp()
    run_app(app)



"""Bluetooth Low Energy Peripheral mocks.
"""

import logging
from typing import Optional, Union, List
from random import randint
from time import sleep

from scapy.packet import Packet

from whad.hub import ProtocolHub
from whad.hub.message import HubMessage
from whad.device.mock import MockDevice
from whad.exceptions import WhadDeviceDisconnected, WhadDeviceNotReady

from whad.hub.ble.address import SetBdAddress
from whad.hub.ble.mode import PeriphMode, BleStart, BleStop
from whad.hub.ble.connect import Connected
from whad.hub.ble import BDAddress, Commands, Direction
from whad.hub.ble.pdu import BlePduReceived, SetAdvData, SendBlePdu, BlePduReceived
from whad.hub.generic.cmdresult import Success, Error, WrongMode
from whad.hub.discovery import Capability, Domain, DeviceType
from whad.ble.profile.attribute import UUID
from whad.ble.profile.characteristic import Characteristic

from .stack.l2cap import LlcapClient

# Create logger for this module.
logger = logging.getLogger(__name__)

class PeripheralMock(MockDevice):
    """BLE peripheral mock device.

    This class implements the intended behavior of a BLE-compatible
    hardware interface.
    """
    STATE_STOPPED = 0
    STATE_STARTED = 1
    STATE_CONNECTED = 2

    def __init__(self, bd_address: str = "aa:bb:cc:dd:ee:ff"):
        """Initialization."""

        # Set state
        self.__state = PeripheralMock.STATE_STOPPED

        # Advertiser address (default: aa:bb:cc:dd:ee:ff)
        self.__bdaddr = BDAddress(bd_address)

        # Advertising data
        self.__adv_data = None
        self.__scan_data = None

        # L2CAP
        self.__l2cap = None
        self.__conn_handle = None

        """Initialization."""
        super().__init__(
            author="Whad Team",
            url="https://whad.io",
            proto_minver=ProtocolHub.LAST_VERSION,
            version="1.0.0",
            dev_type=DeviceType.VirtualDevice,
            dev_id=b"PeripheralMock",
            max_speed=115200,
            capabilities=self.__build_capabilities()
        )

    def __build_capabilities(self) -> dict:
        """Dynamically build the device's capabilities based on its config."""
        # Default commands
        commands = [
            Commands.PeripheralMode,
            Commands.Start, Commands.Stop,
            Commands.SetAdvData,
            Commands.SetBdAddress,
            Commands.SendPDU,
        ]
        capabilities = Capability.NoRawData | Capability.SimulateRole

        return {
            Domain.BtLE : (
                # We can only scan and sniff with no raw data support
                capabilities,
                commands
            )
        }

    def get_adv_data(self) -> Optional[bytes]:
        """Advertising data"""
        return self.__adv_data

    def get_scan_resp(self) -> Optional[bytes]:
        """Scan response data."""
        return self.__scan_data

    def is_started(self) -> bool:
        """Check if peripheral is started."""
        return self.__state == PeripheralMock.STATE_STARTED

    def is_stopped(self) -> bool:
        """Check if peripheral mode is stopped."""
        return self.__state == PeripheralMock.STATE_STOPPED

    def make_connection(self, initiator: BDAddress) -> bool:
        """Initiate a connection from an emulated central device.

        Peripheral must be started and advertising, the mocked interface
        will generate a Connected message and send it to the connector.
        """
        if self.__state == PeripheralMock.STATE_STARTED:
            # Set state as Connected and set a connection handle
            self.__state = PeripheralMock.STATE_CONNECTED
            self.__conn_handle = randint(1, 100)

            # Add a Connected message to the connector received message queue
            msg = Connected(
                initiator=initiator.value,
                advertiser=self.__bdaddr.value,
                access_address=0,
                adv_addr_type=self.__bdaddr.type,
                init_addr_type=initiator.type,
                conn_handle = self.__conn_handle,
            )
            self.put_message(msg)

            # Initialize an emulated L2CAP connection
            self.__l2cap = LlcapClient(self.__conn_handle)
            self.__client = self.__l2cap.get_gatt()

            # Success
            return True
        else:
            # Nope
            return False

    def to_messages(self, pdus: List[Packet]) -> List[HubMessage]:
        """Convert L2CAP PDUs to a list of BlePduReceived messages."""
        messages = []
        for pdu in pdus:
            messages.append(BlePduReceived(
                conn_handle=self.__conn_handle,
                direction=Direction.MASTER_TO_SLAVE,
                pdu=bytes(pdu),
                processed=False,
                decrypted=False
            ))
        return messages

    def wait_procedure(self, timeout: float = 1.0):
        """Retrieve waiting PDUs from L2CAP layer and convert them to messages and send them to the
        attached connector, then wait for the current procedure to complete."""
        if self.__l2cap is None:
            logger.warning("[PeripheralMock] wait_procedure(): L2CAP layer has not been instantiated.")
            return None
        elif self.__client is None:
            logger.warning("[PeripheralMock] wait_procedure(): GATT client object not instantiated.")
        else:
            messages = self.to_messages(self.__l2cap.get_pdus())
            for msg in messages:
                self.put_message(msg)

            # Wait for the client procedure to terminate
            return self.__client.wait_procedure(timeout=timeout)

    def read_by_group_type(self, group_uuid: UUID, start_handle: int, end_handle: int):
        """Emulate a ReadGroupByType procedure initiated by a remote central.

        This method must be called by the application thread to avoid blocking one
        of the mock device's message thread.
        """
        # Start a ReadGroupByType procedure from an emulated central device."""
        self.__client.read_by_group_type(group_uuid, start_handle, end_handle)
        return self.wait_procedure()

    def read_by_type(self, start_handle: int, end_handle: int, type_uuid: UUID):
        """Emulate a ReadByType procedure initiated by a remote central."""
        # Start a ReadByType procedure from an emulated central device.
        self.__client.read_by_type(start_handle, end_handle, type_uuid)
        return self.wait_procedure()

    def read_attr(self, handle: int):
        """Emulate a Read procedure initiated by a remote central."""
        # Start a Read procedure from an emulated central device.
        self.__client.read(handle)
        return self.wait_procedure()

    def read_blob(self, handle: int, offset: int):
        """Emulate a ReadBlob procedure initiated by a remote central."""
        # Start a ReadBlob procedure from an emulated central device.
        self.__client.read_blob(handle, offset)
        return self.wait_procedure()

    def write_attr(self, handle: int, value: bytes):
        """Emulate a Write procedure initiated by a remote central."""
        # Start a Write procedure from an emulated central device.
        self.__client.write(handle, value)
        return self.wait_procedure()

    def write_cmd(self, handle: int, value: bytes):
        """Emulate a WriteCommand procedure initiated by a remote central."""
        # Start a Write procedure from an emulated central device.
        self.__client.write_cmd(handle, value)
        return self.wait_procedure()

    def sub_notif(self, handle: int, charac: Characteristic):
        """Subscribe for notification by writing into the specified attribute (descriptor)."""
        # Start a NotificationCheck procedure
        self.__client.sub_notif(handle)

        if self.__l2cap is None:
            logger.warning("[PeripheralMock] wait_procedure(): L2CAP layer has not been instantiated.")
            return None
        elif self.__client is None:
            logger.warning("[PeripheralMock] wait_procedure(): GATT client object not instantiated.")
        else:
            messages = self.to_messages(self.__l2cap.get_pdus())
            for msg in messages:
                self.put_message(msg)

        sleep(1)
        charac.value = b"FOOBAR"

        # Wait for the client procedure to terminate
        return self.__client.wait_procedure(timeout=2.0)

    def sub_ind(self, handle: int, charac: Characteristic):
        """Subscribe for notification by writing into the specified attribute (descriptor)."""
        # Start a NotificationCheck procedure
        self.__client.sub_ind(handle)

        if self.__l2cap is None:
            logger.warning("[PeripheralMock] wait_procedure(): L2CAP layer has not been instantiated.")
            return None
        elif self.__client is None:
            logger.warning("[PeripheralMock] wait_procedure(): GATT client object not instantiated.")
        else:
            messages = self.to_messages(self.__l2cap.get_pdus())
            for msg in messages:
                self.put_message(msg)

        sleep(1)
        charac.value = b"FOOBAR"

        # Wait for the client procedure to terminate
        return self.__client.wait_procedure(timeout=2.0)

    def find_information(self, start_handle: int, end_handle: int):
        """Emulate a FindInformation procedure initiated by a remote central.

        This method must be called by the application thread to avoid blocking
        one of the mock device's message thread.
        """
        # Start a FindInformation procedure from an emulated central device.
        self.__client.find_information(start_handle, end_handle)
        return self.wait_procedure()

    def find_by_type_value(self, start_handle: int, end_handle: int, attr_type: UUID, attr_value: bytes):
        """Emulate a FindByType procedure initiated by a remote central."""
        # Start a FindInformation procedure from an emulated central device.
        self.__client.find_by_type_value(start_handle, end_handle, attr_type, attr_value)
        return self.wait_procedure()

    @MockDevice.route(PeriphMode)
    def on_periph_mode(self, message: PeriphMode):
        """BLE Peripheral node handler.

        If a mode is already selected and running, return a WrongMode
        error. If in stopped state, return success.
        """
        if self.__state == PeripheralMock.STATE_STARTED:
            return WrongMode()

        # Save advertising parameters
        self.__adv_data = message.get_adv_data()
        self.__scan_data = message.get_scan_data()

        # Success
        return  Success()

    @MockDevice.route(BleStart)
    def on_start(self, _: BleStart):
        """Start selected mode."""
        self.__state = PeripheralMock.STATE_STARTED
        return Success()

    @MockDevice.route(BleStop)
    def on_stop(self, _: BleStop):
        """Stop selected mode."""
        if self.__state == PeripheralMock.STATE_STARTED:
            self.__state = PeripheralMock.STATE_STOPPED
            return Success()
        else:
            return Error()

    @MockDevice.route(SetAdvData)
    def on_set_adv_data(self, _: SetAdvData):
        """Handle SetAdvData message."""
        return Success()

    @MockDevice.route(SetBdAddress)
    def on_set_bd_address(self, _: SetBdAddress):
        """Handle SetBdAddress."""
        return Success()

    @MockDevice.route(SendBlePdu)
    def on_send_pdu(self, send_pdu: SendBlePdu) -> Union[HubMessage, List[HubMessage]]:
        """Handle SendPdu command from connector.

        This callback forwards the given PDU to our emulated central.
        """
        if send_pdu.conn_handle == self.__conn_handle:
            try:
                # Forward to l2cap
                answers = self.__l2cap.on_pdu(send_pdu.to_packet())
                result: List[HubMessage] = [Success()]

                # Convert response PDUs into BlePduReceived messages and add them
                # to the messages sent back to the connector
                for answer in answers:
                    result.append(BlePduReceived(
                        conn_handle=send_pdu.conn_handle,
                        direction=Direction.MASTER_TO_SLAVE,
                        pdu=bytes(answer),
                        processed=False,
                        decrypted=False
                    ))

                # Send messages to connector
                return result
            except (WhadDeviceDisconnected, WhadDeviceNotReady):
                return Error()
        else:
            return Error()




"""BLE GATT Central tool interactive shell.
"""
import re
import html

from typing import List

from prompt_toolkit import print_formatted_text, HTML
from hexdump import hexdump

# We need to import all ATT/L2CAP layers that may be used in python code evaluation
# through `spdu` shell command.
# pylint: disable-next=wildcard-import,unused-wildcard-import
from scapy.layers.bluetooth import *

# We need to import all BTLE layers that may be used in python code evaluation
# through `spdu` shell command.
# pylint: disable-next=wildcard-import,unused-wildcard-import
from scapy.layers.bluetooth4LE import *

from whad.ble.exceptions import InvalidHandleValueException, ConnectionLostException, \
    PeripheralNotFound
from whad.exceptions import ExternalToolNotFound
from whad.device import WhadDevice, WhadDeviceConnector
from whad.ble import Scanner, Central
from whad.ble.profile.device import PeripheralCharacteristic, PeripheralDevice
from whad.ble.profile.characteristic import CharacteristicProperties
from whad.ble.profile.advdata import AdvDataFieldList
from whad.ble.profile.attribute import UUID, InvalidUUIDException
from whad.ble.stack.att.exceptions import AttError, AttributeNotFoundError, \
    InsufficientAuthenticationError, InsufficientAuthorizationError, \
    InsufficientEncryptionKeySize, ReadNotPermittedError, \
    WriteNotPermittedError
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.cli.central.cache import BleDevicesCache
from whad.ble.scanning import AdvertisingDevice
from whad.common.monitors import WiresharkMonitor
from whad.device.unix import UnixSocket

from whad.cli.shell import InteractiveShell, category

INTRO='''
wble-central, the WHAD Bluetooth Low Energy central utility
'''

BDADDR_REGEXP = "^([a-fA-F0-9]{2}:){5}[a-fA-F0-9]{2}$"

def show_adv_record(offset, raw_record):
    """Display advertising record as hexdump.

    @param offset int: AD record offset in AD records data
    @param raw_record bytes: Raw AD record
    """
    nlines = int(len(raw_record)/16)
    if nlines*16 < len(raw_record):
        nlines += 1

    print(f" AD Record #{offset:d}:")
    for line in range(0, nlines):
        line_str = " ".join([f"{c:02x}" for c in raw_record[line*16:(line+1)*16]])
        print("  " + line_str)

class BleCentralShell(InteractiveShell):
    """Bluetooth Low Energy interactive shell
    """

    def __init__(self, interface: WhadDevice = None, connector=None, bd_address=None):
        super().__init__(HTML("<b>wble-central></b>"))

        # If interface is None, pick the first matching our needs
        self.__interface = interface
        self.__cache = BleDevicesCache()
        self.__wireshark = None

        # If connector is not provided
        if connector is None:
            # Reset target info and connector.
            self.__target = None
            self.__target_bd = None
            self.__connector: WhadDeviceConnector = None
        else:
            # If connector provided, consider the central already connected
            self.__connector = connector
            self.__target = connector.peripheral()
            self.__target_bd = bd_address

            # Attach our disconnection callback
            self.__target.set_disconnect_cb(self.on_disconnect)

            # Detach any previous callback
            self.__connector.detach_callback(self.on_disconnect, on_reception=True,\
                on_transmission=False)

            # Attach our packet monitor callback (to detect disconnection)
            self.__connector.attach_callback(
                self.on_disconnect,
                on_transmission=False,
                on_reception=True,
                packet=lambda pkt: pkt.haslayer(LL_TERMINATE_IND)
            )

            # Create our cached device if non-existing
            if self.__target_bd not in self.__cache:
                self.__cache.add(AdvertisingDevice(
                    -50,
                    0,
                    self.__target_bd,
                    AdvDataFieldList()
                ))

        self.intro = INTRO

        self.update_prompt()

    def is_stdin_piped(self) -> bool:
        """Check if we are using a Unix socket as our main device interface.
        """
        return isinstance(self.__interface, UnixSocket)

    def update_prompt(self, force=False):
        """Update prompt to reflect current state
        """
        if not self.__target_bd:
            self.set_prompt(HTML("<b>wble-central></b> "), force)
        else:
            self.set_prompt(HTML(
                f"<b>wble-central|<ansicyan>{self.__target_bd}</ansicyan>></b> "
            ), force)


    def switch_role(self, new_role):
        """Switch from current role (if any) to another role.
        """
        if self.__connector is not None:
            self.__connector.stop()
            if self.__wireshark is not None:
                self.__wireshark.detach()
        self.__connector = new_role(self.__interface)
        if self.__wireshark is not None:
            self.__wireshark.attach(self.__connector)
            self.__wireshark.start()

    def show_att_error(self, error: AttError):
        """Parse ATT error and show exception.
        """
        if isinstance(error, InvalidHandleValueException):
            self.error("ATT Error: wrong value handle")
        elif isinstance(error, ReadNotPermittedError):
            self.error("ATT error: read operation not allowed")
        elif isinstance(error, WriteNotPermittedError):
            self.error("ATT error: write operation not allowed")
        elif isinstance(error, InsufficientAuthenticationError):
            self.error("ATT error: insufficient authentication")
        elif isinstance(error, InsufficientAuthorizationError):
            self.error("ATT error: insufficient authorization")
        elif isinstance(error, AttributeNotFoundError):
            self.error("ATT error: attribute not found")
        elif isinstance(error, InsufficientEncryptionKeySize):
            self.error("ATT error: insufficient encryption")

    @category('Devices discovery')
    def do_clear(self, _):
        """clear cached devices

        <ansicyan><b>clear</b></ansicyan>

        Clear all discovered devices from cache.
        """
        print("Clearing device cache ...")

        # Clear scanner cache as well, if current connector is a scanner.
        if isinstance(self.__connector, Scanner):
            self.__connector.clear()

        # Clear device cache
        self.__cache.clear()


    @category("Devices discovery")
    def do_scan(self, args):
        """scan surrounding devices and show a small summary

        <ansicyan><b>scan</b> <i>[min RSSI (dBm)]</i></ansicyan>

        Scan devices and report them in this console in real-time.

        The following information is provided:
         - <b>RSSI:</b> represents the strength of the signal in dBm
         - <b>Address type:</b> address is either public or random
         - <b>BD address:</b> device Bluetooth address (a.k.a MAC address)
         - <b>Extra info:</b> device name is provided in advertising packet

        You can stop a scan by hitting <b>CTL-c</b> at any time, the discovered devices are kept in
        memory and would be available in autocompletion.
        """
        try:
            # Parse min RSSI level
            rssi_min = -200
            try:
                if len(args) >= 1:
                    rssi_min = int(args[0])
            except ValueError:
                pass

            # Switch role to scanner
            self.switch_role(Scanner)

            # Start scanning
            print_formatted_text(HTML(
                "<ansigreen> RSSI Lvl  Type  BD Address        Extra info</ansigreen>"
            ))
            self.__connector.start()
            try:
                for device in self.__connector.discover_devices():
                    if device.rssi >= rssi_min:
                        # Show device if RSSI is above minimal level
                        print(device)

                    # Add device to cache
                    self.__cache.add(device)
            except KeyboardInterrupt:
                print("\rScan terminated by user")

            # Detach wireshark if needed
            if self.__wireshark is not None:
                self.__wireshark.detach()

            # Stop connector
            self.__connector.stop()

        # Permission error
        except PermissionError:
            interface = self.__interface.interface
            self.error(f"No permission to access the required interface &lt;{interface}&gt; !")

    @category('Devices discovery')
    def do_devices(self, _):
        """list discovered devices

        <ansicyan><b>devices</b></ansicyan>

        List every discovered device so far, through the <ansicyan>scan</ansicyan> command.
        This command displays the content of the console device cache.
        """
        if len(self.__cache) > 0:
            print_formatted_text(HTML(
                "<ansigreen> RSSI Lvl  Type  BD Address        Extra info</ansigreen>"
            ))
            for device in self.__cache.iterate():
                print(device['info'])
        else:
            print("No discovered device in cache.")

    @category('Devices discovery')
    def do_info(self, args):
        """show detailed device information

        <ansicyan><b>info</b> <i>[ BD address ]</i></ansicyan>

        This command displays all the information discovered when scanning a given device:
         - <b>RSSI:</b> represents the strength of the signal in dBm
         - <b>Address type:</b> address is either public or random
         - <b>BD address:</b> device Bluetooth address (a.k.a MAC address)
         - <b>All the advertising records</b> returned in advertising packets and scan responses
        """
        if len(args) >= 1:
            address = args[0]
            try:
                # Retrieve device from cache
                device = self.__cache[address]

                # Show detailed information about the selected device
                dev_info = device['info']
                address_type = 'public' if dev_info.address_type == 0 else 'random'

                print_formatted_text(HTML(
                    f"<ansigreen><b>Device {dev_info.address}</b></ansigreen>"
                ))

                print_formatted_text(HTML(f"<b>RSSI:</b>\t\t\t{dev_info.rssi:4d} dBm"))
                print_formatted_text(HTML(f"<b>Address type:</b>\t\t{address_type}"))
                print('')
                print_formatted_text(HTML("<ansicyan><u>Raw advertising records</u></ansicyan>\n"))
                offset = 0
                for adv_record in dev_info.ad_records:
                    show_adv_record(offset, adv_record.to_bytes())
                    offset += 1
                    print('')
            except IndexError:
                print("!!! Specified BD address has not been discovered")
        else:
            self.error("<u>info</u> requires a single parameter (device name or BD address).")

    def get_cache_targets(self) -> List:
        """Retrieve targets from cache
        """
        # Keep track of BD addresses and names
        targets = [dev['info'].address for dev in self.__cache.iterate()]
        targets.extend([f"\"{dev['info'].name}\"" for dev in self.__cache.iterate()
                        if dev['info'].name is not None])
        return targets

    def complete_info(self):
        """Autocomplete the 'info' command, providing bd addresses of discovered devices.
        """
        # Keep track of BD addresses and names
        targets = self.get_cache_targets()
        completions = self.autocomplete_env(BDADDR_REGEXP)
        for address in targets:
            completions[address] = None
        return completions


    def complete_connect(self):
        """Auto-complete callback for `connect` shell command.
        
        This callback will look into cached devices and populate the suggestions
        with it.
        """
        # Keep track of BD addresses and names
        targets = self.get_cache_targets()
        completions = self.autocomplete_env()
        for address in targets:
            completions[address] = None
        return completions


    @category('GATT client')
    def do_connect(self, args):
        """connect to a device

        <ansicyan><b>connect</b> <i>[ BD address or device name ]</i> <i>[ Connection type ]</i></ansicyan>

        Initiate a Bluetooth Low Energy connection to a specific device by its
        Bluetooth Device address or its name. If multiple devices have the same
        name, the first one will be picked for connect.

        By default, the connection is established with a public connection type if
        the target device is not cached, otherwise it is automatically .
        You can indicate a specific connection type using a second parameter and
        indicating "random" or "public".
        """
        # Check arguments
        if len(args) < 1:
            self.error((
                "<u>connect</u> requires at least one parameter (device name or BD address).\n"
                "type \'help connect\' for more details."))
            return

        # Ensure we are not using a unix socket interface (we cannot connect if this is the case)
        if self.is_stdin_piped():
            self.error(
                "<u>connect</u> cannot be used when wble-central is chained with another tool."
            )
            return

        try:
            target_random_address_type = False
            try:
                target = self.__cache[args[0]]
                target_bd_addr = target['info'].address
                target_random_address_type = target['info'].address_type == 1
            except IndexError:
                # If target not in cache, we are expecting a BD address
                if re.match(BDADDR_REGEXP, args[0]):
                    target_bd_addr = args[0]
                else:
                    self.error("You must provide a valid BD address.")
                    return

            if len(args) == 2:
                if args[1].lower() in ("rnd", "random"):
                    target_random_address_type = True
                elif args[1].lower() in ("pub", "public"):
                    target_random_address_type = False
                else:
                    self.error(
                        "You must indicate a valid connection type ('public' or 'random')."
                    )
                    return

            # Switch role to Central
            self.switch_role(Central)
            try:
                # Try to connect to our target device (central role is started here)
                self.__target = self.__connector.connect(target_bd_addr,
                                                         random=target_random_address_type)
                self.__target_bd = target_bd_addr

                # Check connection is OK
                if self.__target is not None:
                    print(f"Successfully connected to target {target_bd_addr}")

                    # Attach our disconnection callback
                    self.__target.set_disconnect_cb(self.on_disconnect)

                    # Attach our wireshark monitor, if any
                    if self.__wireshark is not None:
                        self.__wireshark.attach(self.__connector)

                    # Detach any previous callback
                    self.__connector.detach_callback(self.on_disconnect, on_reception=True,
                                                     on_transmission=False)

                    # Attach our packet monitor callback (to detect disconnection)
                    self.__connector.attach_callback(
                        self.on_disconnect,
                        on_transmission=False,
                        on_reception=True,
                        packet=lambda pkt: pkt.haslayer(LL_TERMINATE_IND)
                    )

                    # Create our cached device if non-existing
                    if self.__target_bd.lower() not in self.__cache:
                        self.__cache.add(AdvertisingDevice(
                            -50,
                            target_random_address_type,
                            self.__target_bd,
                            AdvDataFieldList()
                        ))

                    # Update prompt
                    self.update_prompt()
                else:
                    print(f"Unable to connect to device {target_bd_addr}")
                    self.__target_bd = None
            except PeripheralNotFound:
                print(f"Unable to connect to device {target_bd_addr}")
                self.__target_bd = None
            except KeyboardInterrupt:
                # Connection interrupted by user, stop current mode
                self.__connector.stop()
        except IndexError:
            print(f"Device {args[0]} not found")
        except PermissionError:
            interface = self.__interface.interface
            self.error(f"No permission to access the required interface &lt;{interface}&gt; !")


    def on_disconnect(self, _=None):
        """Disconnection callback

        This callback is called when a BLE peripheral disconnects our central.
        """
        # Process disconnection
        self.__target_bd = None
        if self.__target is not None:

            # detach wireshark
            if self.__wireshark is not None:
                self.__wireshark.detach()

            self.__target = None

        # Update prompt
        self.update_prompt(force=True)

        # Show disconnection
        print_formatted_text(HTML("<ansired>Peripheral has just disconnected</ansired>"))


    @category("GATT client")
    def do_disconnect(self, _):
        """disconnect from device

        <ansicyan><b>disconnect</b></ansicyan>

        Disconnect from current connected device, if any.
        """
        if self.__target is not None:
            self.__target.disconnect()
            self.__target_bd = None

            # detach wireshark
            if self.__wireshark is not None:
                self.__wireshark.detach()
        else:
            self.warning("not connected to a device, aborted.")

        # Update prompt
        self.update_prompt()


    @category("GATT client")
    def do_profile(self, args):
        """discover device services and characteristics

        <ansicyan><b>profile</b> <i>[PROFILE]</i></ansicyan>

        This command performs a GATT services and characteristics discovery,
        collecting all this information and keeping it in a dedicated <b>cache</b>.

        If <ansicyan>PROFILE</ansicyan> is provided, <i>wble-central</i> will load
        this file JSON content as current device GATT profile instead of discovering
        GATT services and characteristics. If not, it will discover the device services
        and characteristics, this operation may take some time to complete.

        This <b>cached information</b> is then used by commands <ansicyan>read</ansicyan>,
        <ansicyan>services</ansicyan>, and <ansicyan>characteristics</ansicyan> to
        speed up the process.

        <aaa fg="orange">Sometimes this discovery process may cause an error and produces
        incomplete information, in this case try again and cross fingers.</aaa>
        """
        if self.__target is not None:
            # If PROFILE is provided, parse json and load device info.
            if len(args) > 0:
                try:
                    # Load file content
                    with open(args[0],'rb') as dev_profile:
                        profile_json = dev_profile.read()

                    # Recreate a PeripheralDevice based on this profile
                    self.__target = PeripheralDevice(
                        central=self.__connector,
                        gatt_client=self.__connector.connection.gatt,
                        conn_handle=self.__connector.connection.conn_handle,
                        from_json=profile_json
                    )

                except IOError:
                    self.error(f"Cannot load profile json file `{args[0]}`")

            # If not provided, discover services and characteristics
            else:
                try:
                    self.__target.discover()
                except GattTimeoutException:
                    self.error("GATT timeout occurred")
                    return
                except ConnectionLostException:
                    self.error("Services discovery failed (peripheral disconnected)")
                    return
                except AttError:
                    self.error("ATT error, stopping discovery")
                    return

            # Cache our target with its discovered services/characteristics
            self.__cache.add_profile(self.__target_bd, self.__target)
            self.__cache.mark_as_discovered(self.__target_bd)

            # Show services and characteristics
            for service in self.__target.services():
                print_formatted_text(HTML(
                    f"<ansigreen><b>Service {service.name}</b></ansigreen>\n"
                ))
                for charac in service.characteristics():
                    properties = charac.properties
                    charac_rights = []
                    if properties & CharacteristicProperties.READ != 0:
                        charac_rights.append('read')
                    if properties & CharacteristicProperties.WRITE != 0:
                        charac_rights.append('write')
                    if properties & CharacteristicProperties.WRITE_WITHOUT_RESPONSE != 0:
                        charac_rights.append('write_without_response')
                    if properties & CharacteristicProperties.INDICATE != 0:
                        charac_rights.append('indicate')
                    if properties & CharacteristicProperties.NOTIFY != 0:
                        charac_rights.append('notify')
                    print_formatted_text(HTML((
                        f" <b>{charac.name}</b> handle: <b>{charac.handle:d}</b>, "
                        f"value handle: <b>{charac.value_handle:d}</b>"
                    )))

                    access_rights = ', '.join(charac_rights)
                    print_formatted_text(HTML(
                        f"  | <ansicyan>access rights:</ansicyan> <b>{access_rights}</b>"
                    ))

                    for desc in charac.descriptors():
                        print_formatted_text(HTML((
                            "  | <ansiblue><b>Descriptor type {name}</b></ansiblue> "
                            "handle: <b>{handle}</b>"
                        )).format(name=desc.name, handle=desc.handle))
                print('')


    @category("GATT client")
    def do_services(self, _):
        """discover/show current device services

        <ansicyan><b>services</b></ansicyan>

        This command shows the discovered primary services of a connected BLE
        device, and <u>will only work if a device is connected</u>.
        """
        if self.__target_bd is not None:

            # Do we need to discover the services ?
            if len(list(self.__target.services())) == 0:
                try:
                    self.__target.discover()
                    self.__cache.mark_as_discovered(self.__target_bd)
                except GattTimeoutException:
                    self.error("GATT timeout occurred")
                    return
                except ConnectionLostException:
                    self.error((
                        "Services/characteristics discovery failed "
                        "(peripheral disconnected)"))
                    return

            # We are connected to a device, list cached services
            for service in self.__target.services():
                print_formatted_text(HTML((
                    f"<ansicyan><b>{service.uuid}</b></ansicyan> "
                    f"start handle: <b>{service.handle:d}</b>, "
                    f"end handle: <b>{service.end_handle:d}</b>"
                )))
        else:
            self.error("No device connected.")


    @category("GATT client")
    def do_characteristics(self, _):
        """discover/show current device characteristics

        <ansicyan><b>characteristics</b></ansicyan>

        This command shows the services and characteristics of the connected
        device, with their handles and properties.
        """
        if self.__target_bd is not None:

            # Do we need to discover the services ?
            if len(list(self.__target.services())) == 0:
                try:
                    self.__target.discover()
                    self.__cache.mark_as_discovered(self.__target_bd)
                except GattTimeoutException:
                    self.error("GATT timeout occurred")
                    return
                except ConnectionLostException:
                    self.error((
                        "Services/characteristics discovery failed "
                        "(peripheral disconnected)"))
                    return


            # We are connected to a device, list cached services
            for service in self.__target.services():
                for charac in service.characteristics():
                    properties = charac.properties
                    charac_rights = []
                    if properties & CharacteristicProperties.READ != 0:
                        charac_rights.append('read')
                    if properties & CharacteristicProperties.WRITE != 0:
                        charac_rights.append('write')
                    if properties & CharacteristicProperties.WRITE_WITHOUT_RESPONSE != 0:
                        charac_rights.append('write_without_response')
                    if properties & CharacteristicProperties.INDICATE != 0:
                        charac_rights.append('indicate')
                    if properties & CharacteristicProperties.NOTIFY != 0:
                        charac_rights.append('notify')
                    print_formatted_text(HTML((
                        f" <b>{charac.uuid}</b> handle: <b>{charac.handle:d}</b>, "
                        f"value handle: <b>{charac.value_handle:d}</b>"
                    )))

                    access_rights = ', '.join(charac_rights)
                    print_formatted_text(HTML(
                        f"  | <ansicyan>access rights:</ansicyan> <b>{access_rights}</b>"
                    ))

                    for desc in charac.descriptors():
                        print_formatted_text(HTML((
                            f"  | <ansiblue><b>Descriptor type {desc.name}</b></ansiblue> "
                            f"handle: <b>{desc.handle:d}</b>"
                        )))
        else:
            self.error("No device connected.")

    @category("GATT client")
    def do_read(self, args):
        """read a GATT attribute

        <ansicyan><b>read</b> <i>[UUID | handle] ([offset])</i></ansicyan>

        Read an attribute identified by its handle, or read the value of a characteristic
        identified by its UUID (if unique). An optional offset can be provided
        to start reading from the specified byte position (it will issue a
        <i>ReadBlob</i> operation).

        Result is displayed as an hexadecimal dump with corresponding ASCII text:

        > read 41
        00000000: 74 68 69 73 20 69 73 20  61 20 74 65 73 74        this is a test

        """
        if self.__target_bd:
            # parse target arguments
            if len(args) == 0:
                self.error((
                    "You must provide at least a characteristic value handle "
                    "or characteristic UUID."))
                return

            handle = None
            offset = None

            # figure out what the handle is
            if args[0].lower().startswith('0x'):
                try:
                    handle = int(args[0].lower(), 16)
                except ValueError:
                    self.error(f"Wrong handle: {args[0]}")
                    return
            else:
                try:
                    handle = int(args[0])
                except ValueError:
                    try:
                        handle = UUID(args[0])
                    except InvalidUUIDException:
                        self.error(f"Wrong UUID: {args[0]}")
                        return
                    except ValueError:
                        self.error(f"Wrong UUID: {args[0]}")
                        return

            # Check offset and length
            if len(args) >= 2:
                try:
                    offset = int(args[1])
                except ValueError:
                    self.error("Wrong offset value, will use 0 instead.")
                    offset = None

            # Perform characteristic read by handle
            if not isinstance(handle, UUID):
                try:
                    value = self.__target.read(handle, offset=offset, long=True)

                    if value is not None:
                        # Display result as hexdump
                        hexdump(value)
                    else:
                        self.error("Characteristic read failed (empty response).")
                except AttError as att_err:
                    self.show_att_error(att_err)
                except GattTimeoutException:
                    self.error("GATT timeout while reading.")
                except ConnectionLostException:
                    self.error("Characteristic read failed (peripheral disconnected).")

            else:
                # Perform discovery if required
                if not self.__cache.is_discovered(self.__target_bd):
                    self.__target.discover()
                    self.__cache.mark_as_discovered(self.__target_bd)

                # Search characteristic from its UUID
                target_charac = self.__target.char(handle)
                if target_charac is not None:
                    try:
                        # Read data
                        if offset is not None:
                            value = target_charac.read(offset=offset, long=True)
                        else:
                            value = target_charac.read(long=True)

                        # Display result as hexdump
                        hexdump(value)

                    except AttError as att_err:
                        self.show_att_error(att_err)
                    except GattTimeoutException:
                        self.error("GATT timeout while reading.")
                    except ConnectionLostException:
                        self.error("Characteristic read failed (peripheral disconnected)")
                else:
                    self.error(f"No characteristic found with UUID {handle}")
        else:
            self.error("No device connected.")

    def perform_write(self, args, without_response=False):
        """Perform attribute/handle characteristic
        """
        # parse target arguments
        if len(args) <2:
            self.error((
                "You must provide at least a characteristic value handle or "
                "characteristic UUID, and a value to write."))
            return

        handle = None

        # figure out what the handle is
        if args[0].lower().startswith('0x'):
            try:
                handle = int(args[0].lower(), 16)
            except ValueError:
                self.error(f"Wrong handle: {args[0]}")
                return
        else:
            try:
                handle = int(args[0])
            except ValueError:
                try:
                    handle = UUID(args[0])
                except InvalidUUIDException:
                    self.error(f"Wrong UUID: {args[0]}")
                    return
                except ValueError:
                    self.error(f"Wrong UUID: {args[0]}")
                    return

        # Do we have hex data ?
        if args[1].lower() == 'hex':
            # Decode hex data
            hex_data = ''.join(args[2:])
            try:
                char_value = bytes.fromhex(hex_data.replace("\t", ""))
            except ValueError:
                self.error("Provided hex value contains non-hex characters.")
                return
        else:
            char_value = args[1]

        if not isinstance(char_value, bytes):
            char_value = bytes(char_value,'utf-8')

        # Perform ATT write by handle
        if not isinstance(handle, UUID):
            try:
                if without_response:
                    self.__target.write_command(handle, char_value)
                else:
                    self.__target.write(handle, char_value)
            except AttError as att_err:
                self.show_att_error(att_err)
            except GattTimeoutException:
                self.error("GATT timeout while writing.")
            except ConnectionLostException:
                self.error("Characteristic write failed (peripheral disconnected)")

        else:
            # Perform discovery if required
            if not self.__cache.is_discovered(self.__target_bd):
                self.__target.discover()
                self.__cache.mark_as_discovered(self.__target_bd)

            # Search characteristic from its UUID
            target_charac = self.__target.char(handle)
            if target_charac is not None:
                try:
                    if without_response:
                        target_charac.write(char_value, without_response=True)
                    else:
                        target_charac.value = char_value
                except AttError as att_err:
                    self.show_att_error(att_err)
                except GattTimeoutException:
                    self.error("GATT timeout while writing.")
                except ConnectionLostException:
                    self.error("Characteristic write failed (peripheral disconnected)")
            else:
                self.error(f"No characteristic found with UUID {handle}")


    @category("GATT client")
    def do_writecmd(self, args):
        """write data to a GATT attribute without waiting for a response.

        <ansicyan><b>writecmd</b> <i>[UUID | handle] [hex [value] | value ]</i></ansicyan>

        Write data to the specified GATT attribute (identified by its handle) or to
        a characteristic value (identified by its UUID, if unique) without waiting
        for a response.

        Data can be provided hex-encoded if prefixed by "hex":

        > writecmd 41 hex 41 42 43

        The command above will write 'ABC' to attribute identified by the handle 41.

        Data can also be provided as text:

        > writecmd 41 ABC
        """
        if self.__target_bd:
            self.perform_write(args, without_response=True)
        else:
            self.error("No device connected.")


    @category("GATT client")
    def do_write(self, args):
        """write data to a a GATT attribute.

        <ansicyan><b>write</b> <i>[UUID | handle] [hex [value] | value ]</i></ansicyan>

        Write data to the specified GATT attribute (identified by its handle) or to
        a characteristic value (identified by its UUID, if unique).

        Data can be provided hex-encoded if prefixed by "hex":

        > write 41 hex 41 42 43

        The command above will write 'ABC' to attribute identified by the handle 41.

        Data can also be provided as text:

        > write 41 ABC

        """
        if self.__target_bd:
            self.perform_write(args, without_response=False)
        else:
            self.error("No device connected.")


    @category("GATT client")
    def do_pdu(self, args):
        """Send raw PDU to a connected device

        <ansicyan><b>pdu</b> <i>[PDU (hex)]</i></ansicyan>

        Send a raw link-layer PDU to the target device:

        > pdu 03 02 02 03

        The command above will send a <i>LL_TERMINATE_IND</i> control PDU to
        the device.

        """
        if self.__target_bd:
            try:
                if len(args) >= 1:
                    hex_value = ''.join(args)
                    raw_pdu = bytes.fromhex(hex_value.replace(" ", ""))
                    res = self.__connector.send_pdu(
                        BTLE_DATA(raw_pdu),
                        conn_handle=self.__target.conn_handle
                    )
                    if not res:
                        self.error("An error occurred while sending PDU.")
                else:
                    self.error('Invalid hex value.')
            except ValueError:
                self.error("Invalid hex value.")
            except ConnectionLostException:
                self.error("Sending PDU failed (peripheral disconnected)")

        else:
            self.error("No device connected.")

    @category("GATT client")
    def do_spdu(self, args):
        """Send a raw PDU built with scapy to a connected device

        <ansicyan><b>spdu</b> <i>[Scapy packet definition]</i></ansicyan>

        Send a raw link-layer PDU built with scapy, in Python.

        > spdu BTLE_CTRL()/LL_TERMINATE_IND(code=3)

        The command above will send a <i>LL_TERMINATE_IND</i> control PDU to
        the device.
        """
        if self.__target_bd:
            try:
                python_code = ' '.join(args)

                # Create a limited scapy global environment
                scapy_env = {
                    # Headers
                    'BTLE_CTRL': BTLE_CTRL,
                    'L2CAP_Hdr': L2CAP_Hdr,
                    'ATT_Hdr': ATT_Hdr,

                    # ATT
                    'ATT_Error_Response': ATT_Error_Response,
                    'ATT_Exchange_MTU_Request': ATT_Exchange_MTU_Request,
                    'ATT_Exchange_MTU_Response': ATT_Exchange_MTU_Response,
                    'ATT_Execute_Write_Request': ATT_Execute_Write_Request,
                    'ATT_Execute_Write_Response':  ATT_Execute_Write_Response,
                    'ATT_Find_By_Type_Value_Request': ATT_Find_By_Type_Value_Request,
                    'ATT_Find_By_Type_Value_Response': ATT_Find_By_Type_Value_Response,
                    'ATT_Find_Information_Request': ATT_Find_Information_Request,
                    'ATT_Find_Information_Response': ATT_Find_Information_Response,
                    'ATT_Handle_Value_Indication': ATT_Handle_Value_Indication,
                    'ATT_Handle_Value_Notification': ATT_Handle_Value_Notification,
                    'ATT_Prepare_Write_Request': ATT_Prepare_Write_Request,
                    'ATT_Prepare_Write_Response': ATT_Prepare_Write_Response,
                    'ATT_Read_Blob_Request': ATT_Read_Blob_Request,
                    'ATT_Read_Blob_Response': ATT_Read_Blob_Response,
                    'ATT_Read_By_Group_Type_Request': ATT_Read_By_Group_Type_Request,
                    'ATT_Read_By_Group_Type_Response': ATT_Read_By_Group_Type_Response,
                    'ATT_Read_By_Type_Request_128bit': ATT_Read_By_Type_Request_128bit,
                    'ATT_Read_By_Type_Request': ATT_Read_By_Type_Request,
                    'ATT_Read_By_Type_Response': ATT_Read_By_Type_Response,
                    'ATT_Read_Multiple_Request': ATT_Read_Multiple_Request,
                    'ATT_Read_Multiple_Response': ATT_Read_Multiple_Response,
                    'ATT_Read_Request': ATT_Read_Request,
                    'ATT_Read_Response': ATT_Read_Response,
                    'ATT_Write_Command': ATT_Write_Command,
                    'ATT_Write_Request': ATT_Write_Request,
                    'ATT_Write_Response': ATT_Write_Response,

                    # BTLE control PDU
                    'LL_TERMINATE_IND': LL_TERMINATE_IND,
                    'LL_PAUSE_ENC_REQ': LL_PAUSE_ENC_REQ,
                    'LL_CHANNEL_MAP_IND': LL_CHANNEL_MAP_IND,
                    'LL_CONNECTION_PARAM_REQ': LL_CONNECTION_PARAM_REQ,
                    'LL_CONNECTION_UPDATE_IND': LL_CONNECTION_UPDATE_IND,
                    'LL_CONNECTION_PARAM_RSP': LL_CONNECTION_PARAM_RSP,
                    'LL_ENC_REQ': LL_ENC_REQ,
                    'LL_ENC_RSP': LL_ENC_RSP,
                    'LL_FEATURE_REQ': LL_FEATURE_REQ,
                    'LL_FEATURE_RSP': LL_FEATURE_RSP,
                    'LL_LENGTH_REQ': LL_LENGTH_REQ,
                    'LL_LENGTH_RSP': LL_LENGTH_RSP,
                    'LL_MIN_USED_CHANNELS_IND': LL_MIN_USED_CHANNELS_IND,
                    'LL_VERSION_IND': LL_VERSION_IND,
                    'LL_PHY_REQ': LL_PHY_REQ,
                    'LL_PHY_RSP': LL_PHY_RSP,
                    'LL_PING_REQ': LL_PING_REQ,
                    'LL_PING_RSP': LL_PING_RSP,
                    'LL_REJECT_IND': LL_REJECT_IND,
                    'LL_REJECT_EXT_IND': LL_REJECT_EXT_IND,
                    'LL_SLAVE_FEATURE_REQ': LL_SLAVE_FEATURE_REQ,
                    'LL_START_ENC_REQ': LL_START_ENC_REQ,
                    'LL_START_ENC_RSP': LL_START_ENC_RSP,
                    'LL_UNKNOWN_RSP': LL_UNKNOWN_RSP,
                }

                # We know what we are doing, disable pylint warning
                # pylint: disable-next=eval-used
                raw_pdu = eval(python_code, scapy_env, {})
                if isinstance(raw_pdu, (BTLE_CTRL, L2CAP_Hdr)):
                    res = self.__connector.send_pdu(
                        BTLE_DATA()/raw_pdu,
                        conn_handle=self.__target.conn_handle
                    )
                    if not res:
                        self.error("An error occurred while sending PDU.")
                else:
                    self.error("Invalid hex value.")

            except ConnectionLostException:
                self.error("Sending PDU failed (peripheral disconnected)")

            except Exception:
                self.error("An error occurred while assembling packet")
        else:
            self.error("No device connected.")


    @category("GATT client")
    def do_sub(self, args):
        """subscribe to a characteristic

        <ansicyan><b>sub</b> <i>[UUID | handle]</i></ansicyan>

        Subscribe to characteristic notifications/indications, based on a GATT
        characteristic handle or UUID (if unique).

        Once subscribed, notifications and indications will be displayed in the
        console.

        > sub 0xFFF4
        """
        if self.__target_bd:
            # Figure out what the handle is
            if args[0].lower().startswith('0x'):
                try:
                    handle = int(args[0].lower(), 16)
                except ValueError:
                    self.error(f"Wrong handle: {args[0]}")
                    return
            else:
                try:
                    handle = int(args[0])
                except ValueError:
                    try:
                        handle = UUID(args[0])
                    except InvalidUUIDException:
                        self.error(f"Wrong UUID: {args[0]}")
                        return
                    except Exception:
                        self.error(f"Wrong UUID: {args[0]}")
                        return

            # If UUID is provided
            if isinstance(handle, UUID):
                target_charac = self.__target.char(handle)
            elif isinstance(handle, int):
                try:
                    target_charac = self.__target.find_object_by_handle(handle)
                except IndexError:
                    target_charac = None
                if not isinstance(target_charac, PeripheralCharacteristic) or target_charac is None:
                    self.error(f"No characteristic found with handle {handle}")
                    return

            def on_charac_notified(_, value, indication):
                if indication:
                    print_formatted_text(HTML((
                        f"<ansimagenta>Indication</ansimagenta> received from "
                        f"characteristic with handle {handle}"
                    )))
                    hexdump(value)
                else:
                    print_formatted_text(HTML((
                        f"<ansimagenta>Notification</ansimagenta> received from "
                        f"characteristic with handle {handle}"
                    )))
                    hexdump(value)

            if target_charac is not None:
                try:
                    if not target_charac.can_notify():
                        if target_charac.can_indicate():
                            result = target_charac.subscribe(
                                indication=True,
                                callback=on_charac_notified
                            )
                            if result:
                                print_formatted_text(HTML((
                                    f"Successfully subscribed to indication for "
                                    f"characteristic {target_charac.uuid}"
                                )))
                            else:
                                print_formatted_text(HTML((
                                    f"An error occurred when subscribing to indication"
                                    f" for characteristic {target_charac.uuid}"
                                )))
                        else:
                            self.error((
                                "Characteristic does not send notification "
                                "nor indication."))
                    else:
                        result = target_charac.subscribe(
                            notification=True,
                            callback=on_charac_notified
                        )
                        if result:
                            print_formatted_text(HTML((
                                f"Successfully subscribed to notification for "
                                f"characteristic {target_charac.uuid}"
                            )))
                        else:
                            print_formatted_text(HTML((
                                f"An error occurred when subscribing to notification "
                                f"for characteristic {target_charac.uuid}"
                            )))
                except AttError as att_err:
                    self.show_att_error(att_err)
                except GattTimeoutException:
                    self.error("GATT timeout while writing.")
                except ConnectionLostException:
                    self.error("Characteristic subscribe failed (peripheral disconnected)")
            else:
                self.error(f"No characteristic found with UUID {handle}")
        else:
            self.error('No device connected.')

    @category("GATT client")
    def do_unsub(self, args):
        """unsubscribe from a characteristic

        <ansicyan><b>sub</b> <i>[UUID | handle]</i></ansicyan>

        Unsubscribe to characteristic notifications/indications, based on a GATT
        characteristic handle or UUID (if unique).

        Once unsubscribed, no more notifications and indications will be displayed in the
        console.

        Example:

        > unsub 41
        """
        if self.__target_bd:
            # Figure out what the handle is
            if args[0].lower().startswith('0x'):
                try:
                    handle = int(args[0].lower(), 16)
                except ValueError:
                    self.error(f"Wrong handle: {args[0]}")
                    return
            else:
                try:
                    handle = int(args[0])
                except ValueError:
                    try:
                        handle = UUID(args[0])
                    except InvalidUUIDException:
                        self.error(f"Wrong UUID: {args[0]}")
                        return
                    except Exception:
                        self.error(f"Wrong handle: {args[0]}")
                        return

            # If UUID is provided
            if isinstance(handle, UUID):
                target_charac = self.__target.char(handle)
            elif isinstance(handle, int):
                try:
                    target_charac = self.__target.find_object_by_handle(handle)
                except IndexError:
                    target_charac = None
                if not isinstance(target_charac, PeripheralCharacteristic) or target_charac is None:
                    self.error(f"No characteristic found with handle {handle}")
                    return

            if target_charac is not None:
                try:
                    target_charac.unsubscribe()
                    print_formatted_text(HTML(
                        f"Successfully unsubscribed from characteristic {target_charac.uuid}"
                    ))
                except AttError as att_err:
                    self.show_att_error(att_err)
                except GattTimeoutException:
                    self.error("GATT timeout.")
                except ConnectionLostException:
                    self.error("Characteristic unscribe failed (peripheral disconnected)")
            else:
                self.error(f"No characteristic found with UUID {handle}")

        else:
            self.error('No device connected.')

    def complete_wireshark(self):
        """Autocomplete wireshark command
        """
        completions = {}
        if self.__wireshark is not None:
            completions['off'] = {}
        else:
            completions['on'] = {}
        return completions


    @category("Monitoring")
    def do_wireshark(self, arg):
        """launch wireshark to monitor packets

        <ansicyan><b>wireshark</b> <i>["on" | "off"]</i></ansicyan>

        This command launches a wireshark that will display all the packets sent
        and received in the active connection.
        """
        if len(arg) >=1:
            enabled = arg[0].lower()=="on"
            if enabled:
                if self.__wireshark is None:
                    try:
                        self.__wireshark = WiresharkMonitor()
                        if self.__connector is not None:
                            self.__wireshark.attach(self.__connector)
                            self.__wireshark.start()
                    except ExternalToolNotFound:
                        self.error(("Cannot launch Wireshark, "
                                    "please make sure it is installed."))
                else:
                    self.error(("Wireshark is already launched, see "
                                "<ansicyan>wireshark off</ansicyan>"))
            else:
                # Detach monitor if any
                if self.__wireshark is not None:
                    self.__wireshark.detach()
                    self.__wireshark.close()
                    self.__wireshark = None
        else:
            self.error("Missing arguments, see <ansicyan>help wireshark</ansicyan>.")


    @category("GATT client")
    def do_mtu(self, arg):
        """set ATT MTU

        <ansicyan><b>mtu</b> <i>[MTU]</i></ansicyan>

        Set ATT MTU to <i>MTU</i>. <i>MTU</i> must be an integer value.
        """
        if self.__target_bd:
            # check MTU
            if len(arg) >= 1:
                # parse mtu
                try:
                    mtu = int(arg[0])
                    try:
                        self.__target.set_mtu(mtu)
                    except AttError as att_err:
                        self.show_att_error(att_err)
                    except GattTimeoutException:
                        self.error("GATT timeout while exchanging new MTU.")
                except ValueError:
                    self.error("Provided MTU is not a valid decimal integer.")
            else:
                self.error("You must provide the MTU parameter (integer value).")
        else:
            self.error("Not connected to a remote device.")


    def do_quit(self, args):
        """close wble-central
        """
        if self.__target_bd is not None:
            self.__target.disconnect()
        if self.__connector is not None:
            self.__connector.stop()
        if self.__interface is not None:
            self.__interface.close()
        return True

    def do_exit(self, arg):
        """alias for <ansicyan>quit</ansicyan>
        """
        return self.do_quit(arg)
